@@ -12,7 +12,7 @@ module Hledger.Data (
                module Hledger.Data.Account,
                module Hledger.Data.AccountName,
                module Hledger.Data.Amount,
-               module Hledger.Data.Commodity,
+               module Hledger.Data.Balancing,
                module Hledger.Data.Dates,
                module Hledger.Data.Journal,
                module Hledger.Data.Json,
@@ -31,10 +31,11 @@ module Hledger.Data (
               )
 where
 
+import Test.Tasty (testGroup)
 import Hledger.Data.Account
 import Hledger.Data.AccountName
 import Hledger.Data.Amount
-import Hledger.Data.Commodity
+import Hledger.Data.Balancing
 import Hledger.Data.Dates
 import Hledger.Data.Journal
 import Hledger.Data.Json
@@ -47,13 +48,14 @@ import Hledger.Data.StringFormat
 import Hledger.Data.Timeclock
 import Hledger.Data.Transaction
 import Hledger.Data.TransactionModifier
-import Hledger.Data.Types
+import Hledger.Data.Types hiding (MixedAmountKey, Mixed)
 import Hledger.Data.Valuation
-import Hledger.Utils.Test
 
-tests_Data = tests "Data" [
+tests_Data = testGroup "Data" [
    tests_AccountName
   ,tests_Amount
+  ,tests_Dates
+  ,tests_Balancing
   ,tests_Journal
   ,tests_Ledger
   ,tests_Posting
