@@ -1,4 +1,6 @@
-{-# LANGUAGE RecordWildCards, OverloadedStrings #-}
+{-# LANGUAGE CPP               #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE RecordWildCards   #-}
 {-|
 
 
@@ -8,32 +10,50 @@ account, and subaccounting-excluding and -including balances.
 -}
 
 module Hledger.Data.Account
-where
-import Data.List (find, sortOn)
-import Data.List.Extra (groupSort, groupOn)
-import Data.Maybe (fromMaybe)
-import Data.Ord (Down(..))
-import qualified Data.Map as M
-import qualified Data.Text as T
-import Safe (headMay, lookupJustDef)
-import Text.Printf
+( nullacct
+, accountsFromPostings
+, accountTree
+, showAccounts
+, showAccountsBoringFlag
+, printAccounts
+, lookupAccount
+, parentAccounts
+, accountsLevels
+, mapAccounts
+, anyAccounts
+, filterAccounts
+, sumAccounts
+, clipAccounts
+, clipAccountsAndAggregate
+, pruneAccounts
+, flattenAccounts
+, accountSetDeclarationInfo
+, sortAccountNamesByDeclaration
+, sortAccountTreeByAmount
+) where
 
-import Hledger.Data.AccountName
+import qualified Data.HashSet as HS
+import qualified Data.HashMap.Strict as HM
+import Data.List (find, foldl', sortOn)
+import Data.List.Extra (groupOn)
+import qualified Data.Map as M
+import Data.Ord (Down(..))
+import Safe (headMay)
+import Text.Printf (printf)
+
+import Hledger.Data.AccountName (expandAccountName, clipOrEllipsifyAccountName)
 import Hledger.Data.Amount
-import Hledger.Data.Posting ()
 import Hledger.Data.Types
-import Hledger.Utils
 
 
 -- deriving instance Show Account
 instance Show Account where
     show Account{..} = printf "Account %s (boring:%s, postings:%d, ebalance:%s, ibalance:%s)"
-                       (T.map colonToUnderscore aname)  -- hide : so pretty-show doesn't break line
+                       aname
                        (if aboring then "y" else "n" :: String)
                        anumpostings
-                       (showMixedAmount aebalance)
-                       (showMixedAmount aibalance)
-      where colonToUnderscore x = if x == ':' then '_' else x
+                       (wbUnpack $ showMixedAmountB noColour aebalance)
+                       (wbUnpack $ showMixedAmountB noColour aibalance)
 
 instance Eq Account where
   (==) a b = aname a == aname b -- quick equality test for speed
@@ -65,12 +85,12 @@ nullacct = Account
 accountsFromPostings :: [Posting] -> [Account]
 accountsFromPostings ps =
   let
-    grouped = groupSort [(paccount p,pamount p) | p <- ps]
-    counted = [(aname, length amts) | (aname, amts) <- grouped]
-    summed =  [(aname, sumStrict amts) | (aname, amts) <- grouped]  -- always non-empty
-    acctstree      = accountTree "root" $ map fst summed
-    acctswithnumps = mapAccounts setnumps    acctstree      where setnumps    a = a{anumpostings=fromMaybe 0 $ lookup (aname a) counted}
-    acctswithebals = mapAccounts setebalance acctswithnumps where setebalance a = a{aebalance=lookupJustDef nullmixedamt (aname a) summed}
+    summed = foldr (\p -> HM.insertWith addAndIncrement (paccount p) (1, pamount p)) mempty ps
+      where addAndIncrement (n, a) (m, b) = (n + m, a `maPlus` b)
+    acctstree      = accountTree "root" $ HM.keys summed
+    acctswithebals = mapAccounts setnumpsebalance acctstree
+      where setnumpsebalance a = a{anumpostings=numps, aebalance=total}
+              where (numps, total) = HM.lookupDefault (0, nullmixedamt) (aname a) summed
     acctswithibals = sumAccounts acctswithebals
     acctswithparents = tieAccountParents acctswithibals
     acctsflattened = flattenAccounts acctswithparents
@@ -90,6 +110,22 @@ accountTree rootname as = nullacct{aname=rootname, asubs=map (uncurry accountTre
        ,asubs=map (uncurry accountTree') $ M.assocs m
        }
 
+-- | An efficient-to-build tree suggested by Cale Gibbard, probably
+-- better than accountNameTreeFrom.
+newtype FastTree a = T (M.Map a (FastTree a))
+  deriving (Show, Eq, Ord)
+
+mergeTrees :: (Ord a) => FastTree a -> FastTree a -> FastTree a
+mergeTrees (T m) (T m') = T (M.unionWith mergeTrees m m')
+
+treeFromPath :: [a] -> FastTree a
+treeFromPath []     = T M.empty
+treeFromPath (x:xs) = T (M.singleton x (treeFromPath xs))
+
+treeFromPaths :: (Ord a) => [[a]] -> FastTree a
+treeFromPaths = foldl' mergeTrees (T M.empty) . map treeFromPath
+
+
 -- | Tie the knot so all subaccounts' parents are set correctly.
 tieAccountParents :: Account -> Account
 tieAccountParents = tie Nothing
@@ -124,7 +160,7 @@ sumAccounts a
   | otherwise      = a{aibalance=ibal, asubs=subs}
   where
     subs = map sumAccounts $ asubs a
-    ibal = sum $ aebalance a : map aibalance subs
+    ibal = maSum $ aebalance a : map aibalance subs
 
 -- | Remove all subaccounts below a certain depth.
 clipAccounts :: Int -> Account -> Account
@@ -141,7 +177,7 @@ clipAccountsAndAggregate Nothing  as = as
 clipAccountsAndAggregate (Just d) as = combined
     where
       clipped  = [a{aname=clipOrEllipsifyAccountName (Just d) $ aname a} | a <- as]
-      combined = [a{aebalance=sum $ map aebalance same}
+      combined = [a{aebalance=maSum $ map aebalance same}
                  | same@(a:_) <- groupOn aname clipped]
 {-
 test cases, assuming d=1:
@@ -205,8 +241,9 @@ sortAccountTreeByAmount :: NormalSign -> Account -> Account
 sortAccountTreeByAmount normalsign = mapAccounts $ \a -> a{asubs=sortSubs $ asubs a}
   where
     sortSubs = case normalsign of
-        NormallyPositive -> sortOn (Down . normaliseMixedAmountSquashPricesForDisplay . aibalance)
-        NormallyNegative -> sortOn (       normaliseMixedAmountSquashPricesForDisplay . aibalance)
+        NormallyPositive -> sortOn (\a -> (Down $ amt a, aname a))
+        NormallyNegative -> sortOn (\a -> (amt a, aname a))
+    amt = mixedAmountStripPrices . aibalance
 
 -- | Add extra info for this account derived from the Journal's
 -- account directives, if any (comment, tags, declaration order..).
@@ -225,14 +262,14 @@ accountSetDeclarationInfo j a@Account{..} =
 --
 sortAccountNamesByDeclaration :: Journal -> Bool -> [AccountName] -> [AccountName]
 sortAccountNamesByDeclaration j keepparents as =
-  (if keepparents then id else filter (`elem` as)) $  -- maybe discard missing parents that were added
-  map aname $                                         -- keep just the names
-  drop 1 $                                            -- drop the root node that was added
-  flattenAccounts $                                   -- convert to an account list
-  sortAccountTreeByDeclaration $                      -- sort by declaration order (and name)
-  mapAccounts (accountSetDeclarationInfo j) $         -- add declaration order info
-  accountTree "root"                                  -- convert to an account tree
-  as
+    (if keepparents then id else filter (`HS.member` HS.fromList as)) $  -- maybe discard missing parents that were added
+    map aname $                                         -- keep just the names
+    drop 1 $                                            -- drop the root node that was added
+    flattenAccounts $                                   -- convert to an account list
+    sortAccountTreeByDeclaration $                      -- sort by declaration order (and name)
+    mapAccounts (accountSetDeclarationInfo j) $         -- add declaration order info
+    accountTree "root"                                  -- convert to an account tree
+    as
 
 -- | Sort each group of siblings in an account tree by declaration order, then account name.
 -- So each group will contain first the declared accounts,
@@ -266,6 +303,6 @@ showAccountsBoringFlag = unlines . map (show . aboring) . flattenAccounts
 
 showAccountDebug a = printf "%-25s %4s %4s %s"
                      (aname a)
-                     (showMixedAmount $ aebalance a)
-                     (showMixedAmount $ aibalance a)
+                     (wbUnpack . showMixedAmountB noColour $ aebalance a)
+                     (wbUnpack . showMixedAmountB noColour $ aibalance a)
                      (if aboring a then "b" else " " :: String)
