@@ -1,6 +1,5 @@
 {-# LANGUAGE NoMonomorphismRestriction #-}
 {-# LANGUAGE OverloadedStrings #-}
-{-# LANGUAGE CPP #-}
 {-|
 
 'AccountName's are strings like @assets:cash:petty@, with multiple
@@ -21,6 +20,15 @@ module Hledger.Data.AccountName (
   ,accountNameToAccountRegexCI
   ,accountNameTreeFrom
   ,accountSummarisedName
+  ,accountNameInferType
+  ,accountNameType
+  ,assetAccountRegex
+  ,cashAccountRegex
+  ,liabilityAccountRegex
+  ,equityAccountRegex
+  ,conversionAccountRegex
+  ,revenueAccountRegex
+  ,expenseAccountRegex
   ,acctsep
   ,acctsepchar
   ,clipAccountName
@@ -37,18 +45,29 @@ module Hledger.Data.AccountName (
   ,subAccountNamesFrom
   ,topAccountNames
   ,unbudgetedAccountName
+  ,accountNamePostingType
+  ,accountNameWithoutPostingType
+  ,accountNameWithPostingType
+  ,joinAccountNames
+  ,concatAccountNames
+  ,accountNameApplyAliases
+  ,accountNameApplyAliasesMemo
   ,tests_AccountName
 )
 where
 
-import Data.List.Extra (nubSort)
+import Control.Applicative ((<|>))
+import Control.Monad (foldM)
+import Data.Foldable (asum, toList)
 import qualified Data.List.NonEmpty as NE
-#if !(MIN_VERSION_base(4,11,0))
-import Data.Semigroup ((<>))
-#endif
+import qualified Data.Map as M
+import Data.MemoUgly (memo)
+import qualified Data.Set as S
 import Data.Text (Text)
 import qualified Data.Text as T
 import Data.Tree (Tree(..))
+import Safe
+import Text.DocLayout (realLength)
 
 import Hledger.Data.Types
 import Hledger.Utils
@@ -84,6 +103,36 @@ accountSummarisedName a
       cs = accountNameComponents a
       a' = accountLeafName a
 
+-- | Regular expressions matching common English top-level account names,
+-- used as a fallback when account types are not declared.
+assetAccountRegex      = toRegexCI' "^assets?(:|$)"
+cashAccountRegex       = toRegexCI' "^assets?(:.+)?:(cash|bank|che(ck|que?)(ing)?|savings?|current)(:|$)"
+liabilityAccountRegex  = toRegexCI' "^(debts?|liabilit(y|ies))(:|$)"
+equityAccountRegex     = toRegexCI' "^equity(:|$)"
+conversionAccountRegex = toRegexCI' "^equity:(trad(e|ing)|conversion)s?(:|$)"
+revenueAccountRegex    = toRegexCI' "^(income|revenue)s?(:|$)"
+expenseAccountRegex    = toRegexCI' "^expenses?(:|$)"
+
+-- | Try to guess an account's type from its name,
+-- matching common English top-level account names.
+accountNameInferType :: AccountName -> Maybe AccountType
+accountNameInferType a
+  | regexMatchText cashAccountRegex       a = Just Cash
+  | regexMatchText assetAccountRegex      a = Just Asset
+  | regexMatchText liabilityAccountRegex  a = Just Liability
+  | regexMatchText conversionAccountRegex a = Just Conversion
+  | regexMatchText equityAccountRegex     a = Just Equity
+  | regexMatchText revenueAccountRegex    a = Just Revenue
+  | regexMatchText expenseAccountRegex    a = Just Expense
+  | otherwise                               = Nothing
+
+-- Extract the 'AccountType' of an 'AccountName' by looking it up in the
+-- provided Map, traversing the parent accounts if necessary. If none of those
+-- work, try 'accountNameInferType'.
+accountNameType :: M.Map AccountName AccountType -> AccountName -> Maybe AccountType
+accountNameType atypes a = asum (map (`M.lookup` atypes) $ a : parentAccountNames a)
+                         <|> accountNameInferType a
+
 accountNameLevel :: AccountName -> Int
 accountNameLevel "" = 0
 accountNameLevel a = T.length (T.filter (==acctsepchar) a) + 1
@@ -93,6 +142,65 @@ accountNameLevel a = T.length (T.filter (==acctsepchar) a) + 1
 unbudgetedAccountName :: T.Text
 unbudgetedAccountName = "<unbudgeted>"
 
+accountNamePostingType :: AccountName -> PostingType
+accountNamePostingType a
+    | T.null a = RegularPosting
+    | T.head a == '[' && T.last a == ']' = BalancedVirtualPosting
+    | T.head a == '(' && T.last a == ')' = VirtualPosting
+    | otherwise = RegularPosting
+
+accountNameWithoutPostingType :: AccountName -> AccountName
+accountNameWithoutPostingType a = case accountNamePostingType a of
+                                    BalancedVirtualPosting -> textUnbracket a
+                                    VirtualPosting -> textUnbracket a
+                                    RegularPosting -> a
+
+accountNameWithPostingType :: PostingType -> AccountName -> AccountName
+accountNameWithPostingType BalancedVirtualPosting = wrap "[" "]" . accountNameWithoutPostingType
+accountNameWithPostingType VirtualPosting         = wrap "(" ")" . accountNameWithoutPostingType
+accountNameWithPostingType RegularPosting         = accountNameWithoutPostingType
+
+-- | Prefix one account name to another, preserving posting type
+-- indicators like concatAccountNames.
+joinAccountNames :: AccountName -> AccountName -> AccountName
+joinAccountNames a b = concatAccountNames $ filter (not . T.null) [a,b]
+
+-- | Join account names into one. If any of them has () or [] posting type
+-- indicators, these (the first type encountered) will also be applied to
+-- the resulting account name.
+concatAccountNames :: [AccountName] -> AccountName
+concatAccountNames as = accountNameWithPostingType t $ T.intercalate ":" $ map accountNameWithoutPostingType as
+    where t = headDef RegularPosting $ filter (/= RegularPosting) $ map accountNamePostingType as
+
+-- | Rewrite an account name using all matching aliases from the given list, in sequence.
+-- Each alias sees the result of applying the previous aliases.
+-- Or, return any error arising from a bad regular expression in the aliases.
+accountNameApplyAliases :: [AccountAlias] -> AccountName -> Either RegexError AccountName
+accountNameApplyAliases aliases a =
+  let (aname,atype) = (accountNameWithoutPostingType a, accountNamePostingType a)
+  in foldM
+     (\acct alias -> dbg6 "result" $ aliasReplace (dbg6 "alias" alias) (dbg6 "account" acct))
+     aname
+     aliases
+     >>= Right . accountNameWithPostingType atype
+
+-- | Memoising version of accountNameApplyAliases, maybe overkill.
+accountNameApplyAliasesMemo :: [AccountAlias] -> AccountName -> Either RegexError AccountName
+accountNameApplyAliasesMemo aliases = memo (accountNameApplyAliases aliases)
+  -- XXX re-test this memoisation
+
+-- aliasMatches :: AccountAlias -> AccountName -> Bool
+-- aliasMatches (BasicAlias old _) a = old `isAccountNamePrefixOf` a
+-- aliasMatches (RegexAlias re  _) a = regexMatchesCI re a
+
+aliasReplace :: AccountAlias -> AccountName -> Either RegexError AccountName
+aliasReplace (BasicAlias old new) a
+  | old `isAccountNamePrefixOf` a || old == a =
+      Right $ new <> T.drop (T.length old) a
+  | otherwise = Right a
+aliasReplace (RegexAlias re repl) a =
+  fmap T.pack . regexReplace re repl $ T.unpack a -- XXX
+
 -- | Remove some number of account name components from the front of the account name.
 -- If the special "<unbudgeted>" top-level account is present, it is preserved and
 -- dropping affects the rest of the account name.
@@ -113,7 +221,7 @@ accountNameDrop n a
 -- ie these plus all their parent accounts up to the root.
 -- Eg: ["a:b:c","d:e"] -> ["a","a:b","a:b:c","d","d:e"]
 expandAccountNames :: [AccountName] -> [AccountName]
-expandAccountNames as = nubSort $ concatMap expandAccountName as
+expandAccountNames = toList . foldMap (S.fromList . expandAccountName)
 
 -- | "a:b:c" -> ["a","a:b","a:b:c"]
 expandAccountName :: AccountName -> [AccountName]
@@ -121,7 +229,7 @@ expandAccountName = map accountNameFromComponents . NE.tail . NE.inits . account
 
 -- | ["a:b:c","d:e"] -> ["a","d"]
 topAccountNames :: [AccountName] -> [AccountName]
-topAccountNames as = [a | a <- expandAccountNames as, accountNameLevel a == 1]
+topAccountNames = filter ((1==) . accountNameLevel) . expandAccountNames
 
 parentAccountName :: AccountName -> AccountName
 parentAccountName = accountNameFromComponents . init . accountNameComponents
@@ -189,7 +297,7 @@ elideAccountName width s
       where
         elideparts :: Int -> [Text] -> [Text] -> [Text]
         elideparts width done ss
-          | textWidth (accountNameFromComponents $ done++ss) <= width = done++ss
+          | realLength (accountNameFromComponents $ done++ss) <= width = done++ss
           | length ss > 1 = elideparts width (done++[textTakeWidth 2 $ head ss]) (tail ss)
           | otherwise = done++ss
 
@@ -208,54 +316,65 @@ clipOrEllipsifyAccountName (Just 0) = const "..."
 clipOrEllipsifyAccountName n        = clipAccountName n
 
 -- | Escape an AccountName for use within a regular expression.
--- >>> putStr $ escapeName "First?!#$*?$(*) !@^#*? %)*!@#"
+-- >>> putStr . T.unpack $ escapeName "First?!#$*?$(*) !@^#*? %)*!@#"
 -- First\?!#\$\*\?\$\(\*\) !@\^#\*\? %\)\*!@#
-escapeName :: AccountName -> String
-escapeName = T.unpack . T.concatMap escapeChar
+escapeName :: AccountName -> Text
+escapeName = T.concatMap escapeChar
   where
     escapeChar c = if c `elem` escapedChars then T.snoc "\\" c else T.singleton c
     escapedChars = ['[', '?', '+', '|', '(', ')', '*', '$', '^', '\\']
 
 -- | Convert an account name to a regular expression matching it and its subaccounts.
 accountNameToAccountRegex :: AccountName -> Regexp
-accountNameToAccountRegex a = toRegex' $ '^' : escapeName a ++ "(:|$)"  -- PARTIAL: Is this safe after escapeName?
+accountNameToAccountRegex a = toRegex' $ "^" <> escapeName a <> "(:|$)"  -- PARTIAL: Is this safe after escapeName?
 
 -- | Convert an account name to a regular expression matching it and its subaccounts,
 -- case insensitively.
 accountNameToAccountRegexCI :: AccountName -> Regexp
-accountNameToAccountRegexCI a = toRegexCI' $ '^' : escapeName a ++ "(:|$)"  -- PARTIAL: Is this safe after escapeName?
+accountNameToAccountRegexCI a = toRegexCI' $ "^" <> escapeName a <> "(:|$)"  -- PARTIAL: Is this safe after escapeName?
 
 -- | Convert an account name to a regular expression matching it but not its subaccounts.
 accountNameToAccountOnlyRegex :: AccountName -> Regexp
-accountNameToAccountOnlyRegex a = toRegex' $ '^' : escapeName a ++ "$" -- PARTIAL: Is this safe after escapeName?
+accountNameToAccountOnlyRegex a = toRegex' $ "^" <> escapeName a <> "$" -- PARTIAL: Is this safe after escapeName?
 
 -- | Convert an account name to a regular expression matching it but not its subaccounts,
 -- case insensitively.
 accountNameToAccountOnlyRegexCI :: AccountName -> Regexp
-accountNameToAccountOnlyRegexCI a = toRegexCI' $ '^' : escapeName a ++ "$" -- PARTIAL: Is this safe after escapeName?
+accountNameToAccountOnlyRegexCI a = toRegexCI' $ "^" <> escapeName a <> "$" -- PARTIAL: Is this safe after escapeName?
 
 -- -- | Does this string look like an exact account-matching regular expression ?
 --isAccountRegex  :: String -> Bool
 --isAccountRegex s = take 1 s == "^" && take 5 (reverse s) == ")$|:("
 
-tests_AccountName = tests "AccountName" [
-   test "accountNameTreeFrom" $ do
+tests_AccountName = testGroup "AccountName" [
+   testCase "accountNameTreeFrom" $ do
     accountNameTreeFrom ["a"]       @?= Node "root" [Node "a" []]
     accountNameTreeFrom ["a","b"]   @?= Node "root" [Node "a" [], Node "b" []]
     accountNameTreeFrom ["a","a:b"] @?= Node "root" [Node "a" [Node "a:b" []]]
     accountNameTreeFrom ["a:b:c"]   @?= Node "root" [Node "a" [Node "a:b" [Node "a:b:c" []]]]
-  ,test "expandAccountNames" $ do
+  ,testCase "expandAccountNames" $ do
     expandAccountNames ["assets:cash","assets:checking","expenses:vacation"] @?=
      ["assets","assets:cash","assets:checking","expenses","expenses:vacation"]
-  ,test "isAccountNamePrefixOf" $ do
+  ,testCase "isAccountNamePrefixOf" $ do
     "assets" `isAccountNamePrefixOf` "assets" @?= False
     "assets" `isAccountNamePrefixOf` "assets:bank" @?= True
     "assets" `isAccountNamePrefixOf` "assets:bank:checking" @?= True
     "my assets" `isAccountNamePrefixOf` "assets:bank" @?= False
-  ,test "isSubAccountNameOf" $ do
+  ,testCase "isSubAccountNameOf" $ do
     "assets" `isSubAccountNameOf` "assets" @?= False
     "assets:bank" `isSubAccountNameOf` "assets" @?= True
     "assets:bank:checking" `isSubAccountNameOf` "assets" @?= False
     "assets:bank" `isSubAccountNameOf` "my assets" @?= False
+  ,testCase "accountNameInferType" $ do
+    accountNameInferType "assets"            @?= Just Asset
+    accountNameInferType "assets:cash"       @?= Just Cash
+    accountNameInferType "assets:A/R"        @?= Just Asset
+    accountNameInferType "liabilities"       @?= Just Liability
+    accountNameInferType "equity"            @?= Just Equity
+    accountNameInferType "equity:conversion" @?= Just Conversion
+    accountNameInferType "expenses"          @?= Just Expense
+    accountNameInferType "revenues"          @?= Just Revenue
+    accountNameInferType "revenue"           @?= Just Revenue
+    accountNameInferType "income"            @?= Just Revenue
  ]
 
