@@ -1,9 +1,10 @@
-{-# LANGUAGE NoMonoLocalBinds #-}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE TypeFamilies #-}
-{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE FlexibleContexts    #-}
+{-# LANGUAGE NoMonoLocalBinds    #-}
+{-# LANGUAGE OverloadedStrings   #-}
+{-# LANGUAGE PackageImports      #-}
 {-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE PackageImports #-}
+{-# LANGUAGE TupleSections       #-}
+{-# LANGUAGE TypeFamilies        #-}
 {-|
 
 Date parsing and utilities for hledger.
@@ -11,7 +12,8 @@ Date parsing and utilities for hledger.
 For date and time values, we use the standard Day and UTCTime types.
 
 A 'SmartDate' is a date which may be partially-specified or relative.
-Eg 2008\/12\/31, but also 2008\/12, 12\/31, tomorrow, last week, next year.
+Eg 2008\/12\/31, but also 2008\/12, 12\/31, tomorrow, last week, next year,
+in 5 days, in -3 quarters.
 We represent these as a triple of strings like (\"2008\",\"12\",\"\"),
 (\"\",\"\",\"tomorrow\"), (\"\",\"last\",\"week\").
 
@@ -65,42 +67,47 @@ module Hledger.Data.Dates (
   latestSpanContaining,
   smartdate,
   splitSpan,
+  spansFromBoundaries,
+  groupByDateSpan,
   fixSmartDate,
   fixSmartDateStr,
   fixSmartDateStrEither,
   fixSmartDateStrEither',
   yearp,
   daysInSpan,
-  maybePeriod,
+
+  tests_Dates
 )
 where
 
-import Prelude ()
-import "base-compat-batteries" Prelude.Compat hiding (fail)
-import qualified "base-compat-batteries" Control.Monad.Fail.Compat as Fail (MonadFail, fail)
+import qualified Control.Monad.Fail as Fail (MonadFail, fail)
 import Control.Applicative (liftA2)
 import Control.Applicative.Permutations
 import Control.Monad (guard, unless)
-import "base-compat-batteries" Data.List.Compat
 import Data.Char (digitToInt, isDigit, ord)
-import Data.Default
+import Data.Default (def)
 import Data.Foldable (asum)
 import Data.Function (on)
-import Data.Maybe
+import Data.Functor (($>))
+import Data.List (elemIndex, group, sort, sortBy)
+import Data.Maybe (catMaybes, fromMaybe, isJust, mapMaybe)
+import Data.Ord (comparing)
 import qualified Data.Set as Set
 import Data.Text (Text)
 import qualified Data.Text as T
 import Data.Time.Format hiding (months)
 import Data.Time.Calendar
-import Data.Time.Calendar.OrdinalDate
-import Data.Time.Clock
-import Data.Time.LocalTime
+    (Day, addDays, addGregorianYearsClip, addGregorianMonthsClip, diffDays,
+     fromGregorian, fromGregorianValid, toGregorian)
+import Data.Time.Calendar.OrdinalDate (fromMondayStartWeek, mondayStartWeek)
+import Data.Time.Clock (UTCTime, diffUTCTime)
+import Data.Time.LocalTime (getZonedTime, localDay, zonedTimeToLocalTime)
 import Safe (headMay, lastMay, maximumMay, minimumMay)
 import Text.Megaparsec
-import Text.Megaparsec.Char
-import Text.Megaparsec.Char.Lexer (decimal)
-import Text.Megaparsec.Custom
-import Text.Printf
+import Text.Megaparsec.Char (char, char', digitChar, string, string')
+import Text.Megaparsec.Char.Lexer (decimal, signed)
+import Text.Megaparsec.Custom (customErrorBundlePretty)
+import Text.Printf (printf)
 
 import Hledger.Data.Types
 import Hledger.Data.Period
@@ -109,20 +116,19 @@ import Hledger.Utils
 
 -- Help ppShow parse and line-wrap DateSpans better in debug output.
 instance Show DateSpan where
-    show s = "DateSpan " ++ showDateSpan s
-    -- show s = "DateSpan \"" ++ showDateSpan s ++ "\"" -- quotes to help pretty-show
+    show s = "DateSpan " ++ T.unpack (showDateSpan s)
 
-showDate :: Day -> String
-showDate = show
+showDate :: Day -> Text
+showDate = T.pack . show
 
 -- | Render a datespan as a display string, abbreviating into a
 -- compact form if possible.
-showDateSpan :: DateSpan -> String
+showDateSpan :: DateSpan -> Text
 showDateSpan = showPeriod . dateSpanAsPeriod
 
 -- | Like showDateSpan, but show month spans as just the abbreviated month name
 -- in the current locale.
-showDateSpanMonthAbbrev :: DateSpan -> String
+showDateSpanMonthAbbrev :: DateSpan -> Text
 showDateSpanMonthAbbrev = showPeriodMonthAbbrev . dateSpanAsPeriod
 
 -- | Get the current local date.
@@ -160,7 +166,7 @@ spanYears (DateSpan ma mb) = mapMaybe (fmap (first3 . toGregorian)) [ma,mb]
 
 -- | Get overall span enclosing multiple sequentially ordered spans.
 spansSpan :: [DateSpan] -> DateSpan
-spansSpan spans = DateSpan (maybe Nothing spanStart $ headMay spans) (maybe Nothing spanEnd $ lastMay spans)
+spansSpan spans = DateSpan (spanStart =<< headMay spans) (spanEnd =<< lastMay spans)
 
 -- | Split a DateSpan into consecutive whole spans of the specified interval
 -- which fully encompass the original span (and a little more when necessary).
@@ -193,7 +199,7 @@ spansSpan spans = DateSpan (maybe Nothing spanStart $ headMay spans) (maybe Noth
 -- [DateSpan 2007-12-02..2008-01-01,DateSpan 2008-01-02..2008-02-01,DateSpan 2008-02-02..2008-03-01,DateSpan 2008-03-02..2008-04-01]
 -- >>> t (WeekdayOfMonth 2 4) 2011 01 01 2011 02 15
 -- [DateSpan 2010-12-09..2011-01-12,DateSpan 2011-01-13..2011-02-09,DateSpan 2011-02-10..2011-03-09]
--- >>> t (DayOfWeek 2) 2011 01 01 2011 01 15
+-- >>> t (DaysOfWeek [2]) 2011 01 01 2011 01 15
 -- [DateSpan 2010-12-28..2011-01-03,DateSpan 2011-01-04..2011-01-10,DateSpan 2011-01-11..2011-01-17]
 -- >>> t (DayOfYear 11 29) 2011 10 01 2011 10 15
 -- [DateSpan 2010-11-29..2011-11-28]
@@ -202,38 +208,47 @@ spansSpan spans = DateSpan (maybe Nothing spanStart $ headMay spans) (maybe Noth
 --
 splitSpan :: Interval -> DateSpan -> [DateSpan]
 splitSpan _ (DateSpan Nothing Nothing) = [DateSpan Nothing Nothing]
-splitSpan _ s | isEmptySpan s = []
-splitSpan NoInterval     s = [s]
-splitSpan (Days n)       s = splitspan startofday     (applyN n nextday)     s
-splitSpan (Weeks n)      s = splitspan startofweek    (applyN n nextweek)    s
-splitSpan (Months n)     s = splitspan startofmonth   (applyN n nextmonth)   s
-splitSpan (Quarters n)   s = splitspan startofquarter (applyN n nextquarter) s
-splitSpan (Years n)      s = splitspan startofyear    (applyN n nextyear)    s
-splitSpan (DayOfMonth n) s = splitspan (nthdayofmonthcontaining n) (nthdayofmonth n . nextmonth) s
-splitSpan (WeekdayOfMonth n wd) s = splitspan (nthweekdayofmonthcontaining n wd) (advancetonthweekday n wd . nextmonth) s
-splitSpan (DayOfWeek n)  s = splitspan (nthdayofweekcontaining n)  (applyN (n-1) nextday . nextweek)  s
-splitSpan (DayOfYear m n) s = splitspan (nthdayofyearcontaining m n) (applyN (n-1) nextday . applyN (m-1) nextmonth . nextyear) s
--- splitSpan (WeekOfYear n)    s = splitspan startofweek    (applyN n nextweek)    s
--- splitSpan (MonthOfYear n)   s = splitspan startofmonth   (applyN n nextmonth)   s
--- splitSpan (QuarterOfYear n) s = splitspan startofquarter (applyN n nextquarter) s
+splitSpan _ ds | isEmptySpan ds = []
+splitSpan _ ds@(DateSpan (Just s) (Just e)) | s == e = [ds]
+splitSpan NoInterval      ds = [ds]
+splitSpan (Days n)        ds = splitspan startofday     addDays n                    ds
+splitSpan (Weeks n)       ds = splitspan startofweek    addDays (7*n)                ds
+splitSpan (Months n)      ds = splitspan startofmonth   addGregorianMonthsClip n     ds
+splitSpan (Quarters n)    ds = splitspan startofquarter addGregorianMonthsClip (3*n) ds
+splitSpan (Years n)       ds = splitspan startofyear    addGregorianYearsClip n      ds
+splitSpan (DayOfMonth n)  ds = splitspan (nthdayofmonthcontaining n)  addGregorianMonthsClip 1 ds
+splitSpan (DayOfYear m n) ds = splitspan (nthdayofyearcontaining m n) addGregorianYearsClip 1 ds
+splitSpan (WeekdayOfMonth n wd) ds = splitspan (nthweekdayofmonthcontaining n wd) advancemonths 1 ds
+  where
+    advancemonths 0 = id
+    advancemonths w = advancetonthweekday n wd . startofmonth . addGregorianMonthsClip w
+splitSpan (DaysOfWeek [])         ds = [ds]
+splitSpan (DaysOfWeek days@(n:_)) ds = spansFromBoundaries e bdrys
+  where
+    (s, e) = dateSpanSplitLimits (nthdayofweekcontaining n) nextday ds
+    bdrys = concatMap (flip map starts . addDays) [0,7..]
+    -- The first representative of each weekday
+    starts = map (\d -> addDays (toInteger $ d - n) $ nthdayofweekcontaining n s) days
 
 -- Split the given span using the provided helper functions:
 -- start is applied to the span's start date to get the first sub-span's start date
--- next is applied to a sub-span's start date to get the next sub-span's start date
-splitspan :: (Day -> Day) -> (Day -> Day) -> DateSpan -> [DateSpan]
-splitspan _ _ (DateSpan Nothing Nothing) = []
-splitspan start next (DateSpan Nothing (Just e)) = splitspan start next (DateSpan (Just $ start e) (Just $ next $ start e))
-splitspan start next (DateSpan (Just s) Nothing) = splitspan start next (DateSpan (Just $ start s) (Just $ next $ start s))
-splitspan start next span@(DateSpan (Just s) (Just e))
-    | s == e = [span]
-    | otherwise = splitspan' start next span
-    where
-      splitspan' start next (DateSpan (Just s) (Just e))
-          | s >= e = []
-          | otherwise = DateSpan (Just subs) (Just sube) : splitspan' start next (DateSpan (Just sube) (Just e))
-          where subs = start s
-                sube = next subs
-      splitspan' _ _ _ = error' "won't happen, avoids warnings"  -- PARTIAL:
+-- addInterval is applied to an integer n (multiplying it by mult) and the span's start date to get the nth sub-span's start date
+splitspan :: (Day -> Day) -> (Integer -> Day -> Day) -> Int -> DateSpan -> [DateSpan]
+splitspan start addInterval mult ds = spansFromBoundaries e bdrys
+  where
+    (s, e) = dateSpanSplitLimits start (addInterval $ toInteger mult) ds
+    bdrys = mapM (addInterval . toInteger) [0,mult..] $ start s
+
+-- | Fill in missing endpoints for calculating 'splitSpan'.
+dateSpanSplitLimits :: (Day -> Day) -> (Day -> Day) -> DateSpan -> (Day, Day)
+dateSpanSplitLimits start _    (DateSpan (Just s) (Just e)) = (start s, e)
+dateSpanSplitLimits start next (DateSpan (Just s) Nothing)  = (start s, next $ start s)
+dateSpanSplitLimits start next (DateSpan Nothing  (Just e)) = (start e, next $ start e)
+dateSpanSplitLimits _     _    (DateSpan Nothing   Nothing) = error "dateSpanSplitLimits: Should not be nulldatespan"  -- PARTIAL: This case should have been handled in splitSpan
+
+-- | Construct a list of 'DateSpan's from a list of boundaries, which fit within a given range.
+spansFromBoundaries :: Day -> [Day] -> [DateSpan]
+spansFromBoundaries e bdrys = zipWith (DateSpan `on` Just) (takeWhile (< e) bdrys) $ drop 1 bdrys
 
 -- | Count the days in a DateSpan, or if it is open-ended return Nothing.
 daysInSpan :: DateSpan -> Maybe Integer
@@ -257,6 +272,22 @@ spanContainsDate (DateSpan (Just b) (Just e)) d = d >= b && d < e
 periodContainsDate :: Period -> Day -> Bool
 periodContainsDate p = spanContainsDate (periodAsDateSpan p)
 
+-- | Group elements based on where they fall in a list of 'DateSpan's without
+-- gaps. The precondition is not checked.
+groupByDateSpan :: Bool -> (a -> Day) -> [DateSpan] -> [a] -> [(DateSpan, [a])]
+groupByDateSpan showempty date colspans =
+      groupByCols colspans
+    . dropWhile (beforeStart . fst)
+    . sortBy (comparing fst)
+    . map (\x -> (date x, x))
+  where
+    groupByCols []     _  = []
+    groupByCols (c:cs) [] = if showempty then (c, []) : groupByCols cs [] else []
+    groupByCols (c:cs) ps = (c, map snd matches) : groupByCols cs later
+      where (matches, later) = span ((spanEnd c >) . Just . fst) ps
+
+    beforeStart = maybe (const True) (>) $ spanStart =<< headMay colspans
+
 -- | Calculate the intersection of a number of datespans.
 spansIntersect [] = nulldatespan
 spansIntersect [d] = d
@@ -339,9 +370,6 @@ parsePeriodExpr' refdate s =
   either (error' . ("failed to parse:" ++) . customErrorBundlePretty) id $  -- PARTIAL:
   parsePeriodExpr refdate s
 
-maybePeriod :: Day -> Text -> Maybe (Interval,DateSpan)
-maybePeriod refdate = either (const Nothing) Just . parsePeriodExpr refdate
-
 -- | Show a DateSpan as a human-readable pseudo-period-expression string.
 -- dateSpanAsText :: DateSpan -> String
 -- dateSpanAsText (DateSpan Nothing Nothing)   = "all"
@@ -362,21 +390,11 @@ spanFromSmartDate refdate sdate = DateSpan (Just b) (Just e)
       (ry,rm,_) = toGregorian refdate
       (b,e) = span sdate
       span :: SmartDate -> (Day,Day)
-      span (SmartRelative This Day)                 = (refdate, nextday refdate)
-      span (SmartRelative Last Day)                 = (prevday refdate, refdate)
-      span (SmartRelative Next Day)                 = (nextday refdate, addDays 2 refdate)
-      span (SmartRelative This Week)                = (thisweek refdate, nextweek refdate)
-      span (SmartRelative Last Week)                = (prevweek refdate, thisweek refdate)
-      span (SmartRelative Next Week)                = (nextweek refdate, startofweek $ addDays 14 refdate)
-      span (SmartRelative This Month)               = (thismonth refdate, nextmonth refdate)
-      span (SmartRelative Last Month)               = (prevmonth refdate, thismonth refdate)
-      span (SmartRelative Next Month)               = (nextmonth refdate, startofmonth $ addGregorianMonthsClip 2 refdate)
-      span (SmartRelative This Quarter)             = (thisquarter refdate, nextquarter refdate)
-      span (SmartRelative Last Quarter)             = (prevquarter refdate, thisquarter refdate)
-      span (SmartRelative Next Quarter)             = (nextquarter refdate, startofquarter $ addGregorianMonthsClip 6 refdate)
-      span (SmartRelative This Year)                = (thisyear refdate, nextyear refdate)
-      span (SmartRelative Last Year)                = (prevyear refdate, thisyear refdate)
-      span (SmartRelative Next Year)                = (nextyear refdate, startofyear $ addGregorianYearsClip 2 refdate)
+      span (SmartRelative n Day)                    = (addDays n refdate, addDays (n+1) refdate)
+      span (SmartRelative n Week)                   = let d = thisweek refdate in (addDays (7*n) d, addDays (7*n+7) d)
+      span (SmartRelative n Month)                  = let d = thismonth refdate in (addGregorianMonthsClip n d, addGregorianMonthsClip (n+1) d)
+      span (SmartRelative n Quarter)                = let d = thisquarter refdate in (addGregorianMonthsClip (3*n) d, addGregorianMonthsClip (3*n+3) d)
+      span (SmartRelative n Year)                   = let d = thisyear refdate in (addGregorianYearsClip n d, addGregorianYearsClip (n+1) d)
       span (SmartAssumeStart y Nothing)             = (startofyear day, nextyear day) where day = fromGregorian y 1 1
       span (SmartAssumeStart y (Just (m, Nothing))) = (startofmonth day, nextmonth day) where day = fromGregorian y m 1
       span (SmartAssumeStart y (Just (m, Just d)))  = (day, nextday day) where day = fromGregorian y m d
@@ -388,13 +406,13 @@ spanFromSmartDate refdate sdate = DateSpan (Just b) (Just e)
 
 -- | Convert a smart date string to an explicit yyyy\/mm\/dd string using
 -- the provided reference date, or raise an error.
-fixSmartDateStr :: Day -> Text -> String
+fixSmartDateStr :: Day -> Text -> Text
 fixSmartDateStr d s =
   either (error' . printf "could not parse date %s %s" (show s) . show) id $  -- PARTIAL:
-  (fixSmartDateStrEither d s :: Either (ParseErrorBundle Text CustomErr) String)
+  (fixSmartDateStrEither d s :: Either (ParseErrorBundle Text CustomErr) Text)
 
 -- | A safe version of fixSmartDateStr.
-fixSmartDateStrEither :: Day -> Text -> Either (ParseErrorBundle Text CustomErr) String
+fixSmartDateStrEither :: Day -> Text -> Either (ParseErrorBundle Text CustomErr) Text
 fixSmartDateStrEither d = fmap showDate . fixSmartDateStrEither' d
 
 fixSmartDateStrEither'
@@ -476,28 +494,28 @@ fixSmartDateStrEither' d s = case parsewith smartdateonly (T.toLower s) of
 -- t "next january"
 -- "2009-01-01"
 --
+-- >>> t "in 5 days"
+-- "2008-12-01"
+-- >>> t "in 7 months"
+-- "2009-06-01"
+-- >>> t "in -2 weeks"
+-- "2008-11-10"
+-- >>> t "1 quarter ago"
+-- "2008-07-01"
+-- >>> t "1 week ahead"
+-- "2008-12-01"
 fixSmartDate :: Day -> SmartDate -> Day
 fixSmartDate refdate = fix
   where
     fix :: SmartDate -> Day
-    fix (SmartRelative This Day)     = refdate
-    fix (SmartRelative Last Day)     = prevday refdate
-    fix (SmartRelative Next Day)     = nextday refdate
-    fix (SmartRelative This Week)    = thisweek refdate
-    fix (SmartRelative Last Week)    = prevweek refdate
-    fix (SmartRelative Next Week)    = nextweek refdate
-    fix (SmartRelative This Month)   = thismonth refdate
-    fix (SmartRelative Last Month)   = prevmonth refdate
-    fix (SmartRelative Next Month)   = nextmonth refdate
-    fix (SmartRelative This Quarter) = thisquarter refdate
-    fix (SmartRelative Last Quarter) = prevquarter refdate
-    fix (SmartRelative Next Quarter) = nextquarter refdate
-    fix (SmartRelative This Year)    = thisyear refdate
-    fix (SmartRelative Last Year)    = prevyear refdate
-    fix (SmartRelative Next Year)    = nextyear refdate
-    fix (SmartAssumeStart y md)      = fromGregorian y (maybe 1 fst md) (fromMaybe 1 $ snd =<< md)
-    fix (SmartFromReference m d)     = fromGregorian ry (fromMaybe rm m) d
-    fix (SmartMonth m)               = fromGregorian ry m 1
+    fix (SmartRelative n Day)     = addDays n refdate
+    fix (SmartRelative n Week)    = addDays (7*n) $ thisweek refdate
+    fix (SmartRelative n Month)   = addGregorianMonthsClip n $ thismonth refdate
+    fix (SmartRelative n Quarter) = addGregorianMonthsClip (3*n) $ thisquarter refdate
+    fix (SmartRelative n Year)    = addGregorianYearsClip n $ thisyear refdate
+    fix (SmartAssumeStart y md)   = fromGregorian y (maybe 1 fst md) (fromMaybe 1 $ snd =<< md)
+    fix (SmartFromReference m d)  = fromGregorian ry (fromMaybe rm m) d
+    fix (SmartMonth m)            = fromGregorian ry m 1
     (ry, rm, _) = toGregorian refdate
 
 prevday :: Day -> Day
@@ -520,8 +538,6 @@ startofmonth day = fromGregorian y m 1 where (y,m,_) = toGregorian day
 nthdayofmonth d day = fromGregorian y m d where (y,m,_) = toGregorian day
 
 thisquarter = startofquarter
-prevquarter = startofquarter . addGregorianMonthsClip (-3)
-nextquarter = startofquarter . addGregorianMonthsClip 3
 startofquarter day = fromGregorian y (firstmonthofquarter m) 1
     where
       (y,m,_) = toGregorian day
@@ -692,6 +708,8 @@ Examples:
 > october, oct                                (start of month in current year)
 > yesterday, today, tomorrow                  (-1, 0, 1 days from today)
 > last/this/next day/week/month/quarter/year  (-1, 0, 1 periods from the current period)
+> in n days/weeks/months/quarters/years       (n periods from the current period)
+> n days/weeks/months/quarters/years ago      (-n periods from the current period)
 > 20181201                                    (8 digit YYYYMMDD with valid year month and day)
 > 201812                                      (6 digit YYYYMM with valid year and month)
 
@@ -729,19 +747,26 @@ Right (SmartAssumeStart 201813012 Nothing)
 smartdate :: TextParser m SmartDate
 smartdate = choice'
   -- XXX maybe obscures date errors ? see ledgerdate
-    [ yyyymmdd, ymd
+    [ relativeP
+    , yyyymmdd, ymd
     , (\(m,d) -> SmartFromReference (Just m) d) <$> md
-    , (SmartFromReference Nothing <$> decimal) >>= failIfInvalidDate
+    , failIfInvalidDate . SmartFromReference Nothing =<< decimal
     , SmartMonth <$> (month <|> mon)
-    , SmartRelative This Day <$ string' "today"
-    , SmartRelative Last Day <$ string' "yesterday"
-    , SmartRelative Next Day <$ string' "tomorrow"
-    , liftA2 SmartRelative (seqP <* skipNonNewlineSpaces) intervalP
+    , SmartRelative 0    Day <$ string' "today"
+    , SmartRelative (-1) Day <$ string' "yesterday"
+    , SmartRelative 1    Day <$ string' "tomorrow"
     ]
   where
-    seqP = choice [This <$ string' "this", Last <$ string' "last", Next <$ string' "next"]
-    intervalP = choice [Day <$ string' "day", Week <$ string' "week", Month <$ string' "month",
-                        Quarter <$ string' "quarter", Year <$ string' "year"]
+    relativeP = do
+        optional $ string' "in" <* skipNonNewlineSpaces
+        num      <- seqP <* skipNonNewlineSpaces
+        interval <- intervalP <* skipNonNewlineSpaces
+        sign     <- choice [negate <$ string' "ago", id <$ string' "ahead", pure id]
+        return $ SmartRelative (sign num) interval
+
+    seqP = choice [ 0 <$ string' "this", -1 <$ string' "last", 1 <$ string' "next", signed skipNonNewlineSpaces decimal ]
+    intervalP = choice [ Day <$ string' "day", Week <$ string' "week", Month <$ string' "month"
+                       , Quarter <$ string' "quarter", Year <$ string' "year" ] <* optional (char' 's')
 
 -- | Like smartdate, but there must be nothing other than whitespace after the date.
 smartdateonly :: TextParser m SmartDate
@@ -761,7 +786,7 @@ validMonth n = n >= 1 && n <= 12
 validDay n = n >= 1 && n <= 31
 
 failIfInvalidDate :: Fail.MonadFail m => SmartDate -> m SmartDate
-failIfInvalidDate s = unless isValid (Fail.fail $ "bad smart date: " ++ show s) *> return s
+failIfInvalidDate s = unless isValid (Fail.fail $ "bad smart date: " ++ show s) $> s
   where isValid = case s of
             SmartAssumeStart y (Just (m, md)) -> isJust $ fromGregorianValid y m (fromMaybe 1 md)
             SmartFromReference mm d           -> isJust $ fromGregorianValid 2004 (fromMaybe 1 mm) d
@@ -777,10 +802,17 @@ yyyymmdd = do
   failIfInvalidDate date
 
 ymd :: TextParser m SmartDate
-ymd = liftA2 SmartAssumeStart yearp (optional $ try monthday) >>= failIfInvalidDate
-  where monthday = do
-            sep <- datesepchar
-            liftA2 (,) decimal . optional $ char sep *> decimal
+ymd = do
+    y <- yearp
+    fmap (SmartAssumeStart y) . optional . try $ do
+        sep <- datesepchar
+        m <- decimal
+        unless (validMonth m) $ Fail.fail ("Bad month " <> show m)
+        fmap (m,) . optional . try $ do
+            _ <- char sep
+            d <- decimal
+            failIfInvalidDate $ SmartAssumeStart y (Just (m, Just d))
+            return d
 
 md :: TextParser m (Month, MonthDay)
 md = do
@@ -817,6 +849,9 @@ weekday = do
     []    -> Fail.fail $ "weekday: should not happen: attempted to find " <>
                          show wday <> " in " <> show (weekdays ++ weekdayabbrevs)
 
+weekdaysp :: TextParser m [Int]
+weekdaysp = fmap head . group . sort <$> sepBy1 weekday (string' ",")
+
 -- | Parse a period expression, specifying a date span and optionally
 -- a reporting interval. Requires a reference "today" date for
 -- resolving any relative start/end dates (only; it is not needed for
@@ -860,9 +895,9 @@ weekday = do
 -- >>> p "every 1st monday of month to 2009"
 -- Right (WeekdayOfMonth 1 1,DateSpan ..2008-12-31)
 -- >>> p "every tue"
--- Right (DayOfWeek 2,DateSpan ..)
+-- Right (DaysOfWeek [2],DateSpan ..)
 -- >>> p "every 2nd day of week"
--- Right (DayOfWeek 2,DateSpan ..)
+-- Right (DaysOfWeek [2],DateSpan ..)
 -- >>> p "every 2nd day of month"
 -- Right (DayOfMonth 2,DateSpan ..)
 -- >>> p "every 2nd day"
@@ -891,7 +926,6 @@ intervalanddateperiodexprp rdate = do
 reportingintervalp :: TextParser m Interval
 reportingintervalp = choice'
     [ tryinterval "day"     "daily"     Days
-    , tryinterval "week"    "weekly"    Weeks
     , tryinterval "month"   "monthly"   Months
     , tryinterval "quarter" "quarterly" Quarters
     , tryinterval "year"    "yearly"    Years
@@ -899,13 +933,20 @@ reportingintervalp = choice'
     , Weeks 2 <$ string' "fortnightly"
     , Months 2 <$ string' "bimonthly"
     , string' "every" *> skipNonNewlineSpaces *> choice'
-        [ DayOfWeek <$> (nth <* skipNonNewlineSpaces <* string' "day" <* of_ "week")
+        [ DaysOfWeek . pure <$> (nth <* skipNonNewlineSpaces <* string' "day" <* of_ "week")
         , DayOfMonth <$> (nth <* skipNonNewlineSpaces <* string' "day" <* optOf_ "month")
         , liftA2 WeekdayOfMonth nth $ skipNonNewlineSpaces *> weekday <* optOf_ "month"
         , uncurry DayOfYear <$> (md <* optOf_ "year")
-        , DayOfWeek <$> weekday
+        , DaysOfWeek <$> weekdaysp
+        , DaysOfWeek [1..5] <$ string' "weekday"
+        , DaysOfWeek [6..7] <$ string' "weekendday"
         , d_o_y <* optOf_ "year"
         ]
+    -- NB: the ordering is important here since the parse for `every weekday`
+    -- would match the `tryinterval` first and then error on `d`. Perhaps it
+    -- would be clearer to factor some of this into the `every` choice or other
+    -- left-factorings.
+    , tryinterval "week"    "weekly"    Weeks
     ]
   where
     of_ period =
@@ -918,18 +959,14 @@ reportingintervalp = choice'
                                               (toPermutation $ nth <* skipNonNewlineSpaces)
 
     -- Parse any of several variants of a basic interval, eg "daily", "every day", "every N days".
-    tryinterval :: String -> String -> (Int -> Interval) -> TextParser m Interval
+    tryinterval :: Text -> Text -> (Int -> Interval) -> TextParser m Interval
     tryinterval singular compact intcons = intcons <$> choice'
-        [ 1 <$ string' compact'
+        [ 1 <$ string' compact
         , string' "every" *> skipNonNewlineSpaces *> choice
-            [ 1 <$ string' singular'
-            , decimal <* skipNonNewlineSpaces <* string' plural'
+            [ 1 <$ string' singular
+            , decimal <* skipNonNewlineSpaces <* string' (singular <> "s")
             ]
         ]
-      where
-        compact'  = T.pack compact
-        singular' = T.pack singular
-        plural'   = T.pack $ singular ++ "s"
 
 periodexprdatespanp :: Day -> TextParser m DateSpan
 periodexprdatespanp rdate = choice $ map try [
@@ -943,6 +980,14 @@ periodexprdatespanp rdate = choice $ map try [
 -- |
 -- >>> parsewith (doubledatespanp (fromGregorian 2018 01 01) <* eof) "20180101-201804"
 -- Right DateSpan 2018Q1
+-- >>> parsewith (doubledatespanp (fromGregorian 2018 01 01) <* eof) "2017..2018"
+-- Right DateSpan 2017
+-- >>> parsewith (doubledatespanp (fromGregorian 2018 01 01) <* eof) "2017-2018"
+-- Right DateSpan 2017
+-- >>> parsewith (doubledatespanp (fromGregorian 2018 01 01) <* eof) "2017-01-2018"
+-- Right DateSpan 2017
+-- >>> parsewith (doubledatespanp (fromGregorian 2018 01 01) <* eof) "2017-01-01-2018"
+-- Right DateSpan 2017
 doubledatespanp :: Day -> TextParser m DateSpan
 doubledatespanp rdate = liftA2 fromToSpan
     (optional (string' "from" *> skipNonNewlineSpaces) *> smartdate)
@@ -994,3 +1039,45 @@ emptydatespan = DateSpan (Just $ addDays 1 nulldate) (Just nulldate)
 
 nulldate :: Day
 nulldate = fromGregorian 0 1 1
+
+
+-- tests
+
+tests_Dates = testGroup "Dates"
+  [ testCase "weekday" $ do
+      splitSpan (DaysOfWeek [1..5]) (DateSpan (Just $ fromGregorian 2021 07 01) (Just $ fromGregorian 2021 07 08))
+        @?= [ (DateSpan (Just $ fromGregorian 2021 06 28) (Just $ fromGregorian 2021 06 29))
+            , (DateSpan (Just $ fromGregorian 2021 06 29) (Just $ fromGregorian 2021 06 30))
+            , (DateSpan (Just $ fromGregorian 2021 06 30) (Just $ fromGregorian 2021 07 01))
+            , (DateSpan (Just $ fromGregorian 2021 07 01) (Just $ fromGregorian 2021 07 02))
+            , (DateSpan (Just $ fromGregorian 2021 07 02) (Just $ fromGregorian 2021 07 05))
+            -- next week
+            , (DateSpan (Just $ fromGregorian 2021 07 05) (Just $ fromGregorian 2021 07 06))
+            , (DateSpan (Just $ fromGregorian 2021 07 06) (Just $ fromGregorian 2021 07 07))
+            , (DateSpan (Just $ fromGregorian 2021 07 07) (Just $ fromGregorian 2021 07 08))
+            ]
+
+      splitSpan (DaysOfWeek [1, 5]) (DateSpan (Just $ fromGregorian 2021 07 01) (Just $ fromGregorian 2021 07 08))
+        @?= [ (DateSpan (Just $ fromGregorian 2021 06 28) (Just $ fromGregorian 2021 07 02))
+            , (DateSpan (Just $ fromGregorian 2021 07 02) (Just $ fromGregorian 2021 07 05))
+            -- next week
+            , (DateSpan (Just $ fromGregorian 2021 07 05) (Just $ fromGregorian 2021 07 09))
+            ]
+
+  , testCase "match dayOfWeek" $ do
+      let dayofweek n s = splitspan (nthdayofweekcontaining n) (\w -> (if w == 0 then id else applyN (n-1) nextday . applyN (fromInteger w) nextweek)) 1 s
+          match ds day = splitSpan (DaysOfWeek [day]) ds @?= dayofweek day ds
+          ys2021 = fromGregorian 2021 01 01
+          ye2021 = fromGregorian 2021 12 31
+          ys2022 = fromGregorian 2022 01 01
+      mapM_ (match (DateSpan (Just ys2021) (Just ye2021))) [1..7]
+      mapM_ (match (DateSpan (Just ys2021) (Just ys2022))) [1..7]
+      mapM_ (match (DateSpan (Just ye2021) (Just ys2022))) [1..7]
+
+      mapM_ (match (DateSpan (Just ye2021) Nothing)) [1..7]
+      mapM_ (match (DateSpan (Just ys2022) Nothing)) [1..7]
+
+      mapM_ (match (DateSpan Nothing (Just ye2021))) [1..7]
+      mapM_ (match (DateSpan Nothing (Just ys2022))) [1..7]
+
+  ]
