@@ -2,28 +2,10 @@
 JSON instances. Should they be in Types.hs ?
 -}
 
-{-# OPTIONS_GHC -fno-warn-orphans #-}
-
-{-# LANGUAGE CPP                 #-}
---{-# LANGUAGE DataKinds           #-}
---{-# LANGUAGE DeriveAnyClass      #-}
-{-# LANGUAGE DeriveGeneric       #-}
---{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE FlexibleInstances   #-}
---{-# LANGUAGE NamedFieldPuns #-}
---{-# LANGUAGE OverloadedStrings   #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE LambdaCase        #-}
 {-# LANGUAGE OverloadedStrings #-}
---{-# LANGUAGE PolyKinds           #-}
---{-# LANGUAGE QuasiQuotes         #-}
---{-# LANGUAGE QuasiQuotes #-}
---{-# LANGUAGE Rank2Types #-}
---{-# LANGUAGE RankNTypes #-}
-{-# LANGUAGE RecordWildCards #-}
---{-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE StandaloneDeriving #-}
---{-# LANGUAGE TemplateHaskell       #-}
---{-# LANGUAGE TypeFamilies        #-}
---{-# LANGUAGE TypeOperators       #-}
+{-# LANGUAGE RecordWildCards   #-}
 
 module Hledger.Data.Json (
   -- * Instances
@@ -33,27 +15,29 @@ module Hledger.Data.Json (
   ,readJsonFile
 ) where
 
-#if !(MIN_VERSION_base(4,13,0))
-import           Data.Semigroup ((<>))
-#endif
 import           Data.Aeson
-import           Data.Aeson.Encode.Pretty (encodePrettyToTextBuilder)
+import           Data.Aeson.Encode.Pretty (Config(..), Indent(..), NumberFormat(..),
+                     encodePretty', encodePrettyToTextBuilder')
 --import           Data.Aeson.TH
 import qualified Data.ByteString.Lazy as BL
-import           Data.Decimal
-import           Data.Maybe
+import           Data.Decimal (DecimalRaw(..), roundTo)
+import           Data.Maybe (fromMaybe)
 import qualified Data.Text.Lazy    as TL
-import qualified Data.Text.Lazy.IO as TL
-import           Data.Text.Lazy.Builder (toLazyText)
-import           GHC.Generics (Generic)
-import           System.Time (ClockTime)
+import qualified Data.Text.Lazy.Builder as TB
+import           Text.Megaparsec (Pos, SourcePos, mkPos, unPos)
 
 import           Hledger.Data.Types
+import           Hledger.Data.Amount (amountsRaw, mixed)
 
 -- To JSON
 
 instance ToJSON Status
-instance ToJSON GenericSourcePos
+instance ToJSON SourcePos
+
+-- Use the same encoding as the underlying Int
+instance ToJSON Pos where
+  toJSON = toJSON . unPos
+  toEncoding = toEncoding . unPos
 
 -- https://github.com/simonmichael/hledger/issues/1195
 
@@ -79,90 +63,126 @@ instance ToJSON GenericSourcePos
 -- and that the overall number of significant digits in the floating point
 -- remains manageable in practice. (I'm not sure how to limit the number
 -- of significant digits in a Decimal right now.)
-instance ToJSON Decimal where
-  toJSON d = object
-    ["decimalPlaces"   .= toJSON decimalPlaces
-    ,"decimalMantissa" .= toJSON decimalMantissa
-    ,"floatingPoint"   .= toJSON (fromRational $ toRational d' :: Double)
+instance (Integral a, ToJSON a) => ToJSON (DecimalRaw a) where
+  toJSON = object . decimalKV
+  toEncoding = pairs . mconcat . decimalKV
+
+decimalKV :: (KeyValue kv, Integral a, ToJSON a) => DecimalRaw a -> [kv]
+decimalKV d = let d' = if decimalPlaces d <= 10 then d else roundTo 10 d in
+    [ "decimalPlaces"   .= decimalPlaces d'
+    , "decimalMantissa" .= decimalMantissa d'
+    , "floatingPoint"   .= (realToFrac d' :: Double)
     ]
-    where d'@Decimal{..} = roundTo 10 d
 
 instance ToJSON Amount
 instance ToJSON AmountStyle
-instance ToJSON AmountPrecision
+
+-- Use the same JSON serialisation as Maybe Word8
+instance ToJSON AmountPrecision where
+  toJSON = toJSON . \case
+    Precision n      -> Just n
+    NaturalPrecision -> Nothing
+  toEncoding = toEncoding . \case
+    Precision n      -> Just n
+    NaturalPrecision -> Nothing
+
 instance ToJSON Side
 instance ToJSON DigitGroupStyle
-instance ToJSON MixedAmount
+
+instance ToJSON MixedAmount where
+  toJSON = toJSON . amountsRaw
+  toEncoding = toEncoding . amountsRaw
+
 instance ToJSON BalanceAssertion
 instance ToJSON AmountPrice
 instance ToJSON MarketPrice
 instance ToJSON PostingType
 
 instance ToJSON Posting where
-  toJSON Posting{..} = object
-    ["pdate"             .= pdate
-    ,"pdate2"            .= pdate2
-    ,"pstatus"           .= pstatus
-    ,"paccount"          .= paccount
-    ,"pamount"           .= pamount
-    ,"pcomment"          .= pcomment
-    ,"ptype"             .= ptype
-    ,"ptags"             .= ptags
-    ,"pbalanceassertion" .= pbalanceassertion
+  toJSON = object . postingKV
+  toEncoding = pairs . mconcat . postingKV
+
+postingKV :: KeyValue kv => Posting -> [kv]
+postingKV Posting{..} =
+    [ "pdate"             .= pdate
+    , "pdate2"            .= pdate2
+    , "pstatus"           .= pstatus
+    , "paccount"          .= paccount
+    , "pamount"           .= pamount
+    , "pcomment"          .= pcomment
+    , "ptype"             .= ptype
+    , "ptags"             .= ptags
+    , "pbalanceassertion" .= pbalanceassertion
     -- To avoid a cycle, show just the parent transaction's index number
     -- in a dummy field. When re-parsed, there will be no parent.
-    ,"ptransaction_"     .= maybe "" (show.tindex) ptransaction
+    , "ptransaction_"     .= maybe "" (show.tindex) ptransaction
     -- This is probably not wanted in json, we discard it.
-    ,"poriginal"         .= (Nothing :: Maybe Posting)
+    , "poriginal"         .= (Nothing :: Maybe Posting)
     ]
 
 instance ToJSON Transaction
 instance ToJSON TransactionModifier
+instance ToJSON TMPostingRule
 instance ToJSON PeriodicTransaction
 instance ToJSON PriceDirective
 instance ToJSON DateSpan
 instance ToJSON Interval
+instance ToJSON Period
 instance ToJSON AccountAlias
 instance ToJSON AccountType
 instance ToJSONKey AccountType
 instance ToJSON AccountDeclarationInfo
+instance ToJSON PayeeDeclarationInfo
 instance ToJSON Commodity
 instance ToJSON TimeclockCode
 instance ToJSON TimeclockEntry
-instance ToJSON ClockTime
 instance ToJSON Journal
 
 instance ToJSON Account where
-  toJSON a = object
-    ["aname"        .= aname a
-    ,"aebalance"    .= aebalance a
-    ,"aibalance"    .= aibalance a
-    ,"anumpostings" .= anumpostings a
-    ,"aboring"      .= aboring a
+  toJSON = object . accountKV
+  toEncoding = pairs . mconcat . accountKV
+
+accountKV :: KeyValue kv => Account -> [kv]
+accountKV a =
+    [ "aname"        .= aname a
+    , "aebalance"    .= aebalance a
+    , "aibalance"    .= aibalance a
+    , "anumpostings" .= anumpostings a
+    , "aboring"      .= aboring a
     -- To avoid a cycle, show just the parent account's name
     -- in a dummy field. When re-parsed, there will be no parent.
-    ,"aparent_"     .= maybe "" aname (aparent a)
+    , "aparent_"     .= maybe "" aname (aparent a)
     -- Just the names of subaccounts, as a dummy field, ignored when parsed.
-    ,"asubs_"       .= map aname (asubs a)
+    , "asubs_"       .= map aname (asubs a)
     -- The actual subaccounts (and their subs..), making a (probably highly redundant) tree
     -- ,"asubs"        .= asubs a
     -- Omit the actual subaccounts
-    ,"asubs"        .= ([]::[Account])
+    , "asubs"        .= ([]::[Account])
     ]
 
-deriving instance Generic (Ledger)
 instance ToJSON Ledger
 
 -- From JSON
 
 instance FromJSON Status
-instance FromJSON GenericSourcePos
+instance FromJSON SourcePos
+-- Use the same encoding as the underlying Int
+instance FromJSON Pos where
+  parseJSON = fmap mkPos . parseJSON
+
 instance FromJSON Amount
 instance FromJSON AmountStyle
-instance FromJSON AmountPrecision
+
+-- Use the same JSON serialisation as Maybe Word8
+instance FromJSON AmountPrecision where
+  parseJSON = fmap (maybe NaturalPrecision Precision) . parseJSON
+
 instance FromJSON Side
 instance FromJSON DigitGroupStyle
-instance FromJSON MixedAmount
+
+instance FromJSON MixedAmount where
+  parseJSON = fmap (mixed :: [Amount] -> MixedAmount) . parseJSON
+
 instance FromJSON BalanceAssertion
 instance FromJSON AmountPrice
 instance FromJSON MarketPrice
@@ -187,9 +207,6 @@ instance FromJSON Account
 -- $(deriveFromJSON defaultOptions ''DecimalRaw)  -- works; requires TH, but gives better parse error messages
 --
 -- https://github.com/PaulJohnson/Haskell-Decimal/issues/6
---deriving instance Generic Decimal
---instance FromJSON Decimal
-deriving instance Generic (DecimalRaw a)
 instance FromJSON (DecimalRaw Integer)
 --
 -- @simonmichael, I think the code in your first comment should work if it compiles—though “work” doesn’t mean you can parse a JSON number directly into a `Decimal` using the generic instance, as you’ve discovered.
@@ -219,6 +236,7 @@ instance FromJSON (DecimalRaw Integer)
 -- instance FromJSON Commodity
 -- instance FromJSON DateSpan
 -- instance FromJSON Interval
+-- instance FromJSON Period
 -- instance FromJSON PeriodicTransaction
 -- instance FromJSON PriceDirective
 -- instance FromJSON TimeclockCode
@@ -229,15 +247,18 @@ instance FromJSON (DecimalRaw Integer)
 
 -- Utilities
 
+-- | Config for pretty printing JSON output.
+jsonConf :: Config
+jsonConf = Config{confIndent=Spaces 2, confCompare=compare, confNumFormat=Generic, confTrailingNewline=True}
+
 -- | Show a JSON-convertible haskell value as pretty-printed JSON text.
 toJsonText :: ToJSON a => a -> TL.Text
-toJsonText = (<>"\n") . toLazyText . encodePrettyToTextBuilder
+toJsonText = TB.toLazyText . encodePrettyToTextBuilder' jsonConf
 
 -- | Write a JSON-convertible haskell value to a pretty-printed JSON file.
 -- Eg: writeJsonFile "a.json" nulltransaction
 writeJsonFile :: ToJSON a => FilePath -> a -> IO ()
-writeJsonFile f = TL.writeFile f . toJsonText
--- we write with Text and read with ByteString, is that fine ?
+writeJsonFile f = BL.writeFile f . encodePretty' jsonConf
 
 -- | Read a JSON file and decode it to the target type, or raise an error if we can't.
 -- Eg: readJsonFile "a.json" :: IO Transaction
@@ -250,4 +271,3 @@ readJsonFile f = do
   case fromJSON v :: FromJSON a => Result a of
     Error e   -> error e
     Success t -> return t
-
