@@ -2,7 +2,10 @@
 -- hledger's report item fields. The formats are used by
 -- report-specific renderers like renderBalanceReportItem.
 
-{-# LANGUAGE FlexibleContexts, OverloadedStrings, TypeFamilies, PackageImports #-}
+{-# LANGUAGE FlexibleContexts  #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE PackageImports    #-}
+{-# LANGUAGE TypeFamilies      #-}
 
 module Hledger.Data.StringFormat (
           parseStringFormat
@@ -10,20 +13,21 @@ module Hledger.Data.StringFormat (
         , StringFormat(..)
         , StringFormatComponent(..)
         , ReportItemField(..)
+        , defaultBalanceLineFormat
         , tests_StringFormat
         ) where
 
-import Prelude ()
-import "base-compat-batteries" Prelude.Compat
-import Numeric
+import Numeric (readDec)
 import Data.Char (isPrint)
-import Data.Maybe
--- import qualified Data.Text as T
+import Data.Default (Default(..))
+import Data.Maybe (isJust)
+import Data.Text (Text)
+import qualified Data.Text as T
 import Text.Megaparsec
-import Text.Megaparsec.Char
+import Text.Megaparsec.Char (char, digitChar, string)
 
-import Hledger.Utils.Parse
-import Hledger.Utils.String (formatString)
+import Hledger.Utils.Parse (SimpleTextParser)
+import Hledger.Utils.Text (formatText)
 import Hledger.Utils.Test
 
 -- | A format specification/template to use when rendering a report line item as text.
@@ -41,13 +45,13 @@ import Hledger.Utils.Test
 -- mode, which provides a limited StringFormat renderer.
 --
 data StringFormat =
-    OneLine [StringFormatComponent]       -- ^ multi-line values will be rendered on one line, comma-separated
-  | TopAligned [StringFormatComponent]    -- ^ values will be top-aligned (and bottom-padded to the same height)
+    OneLine       [StringFormatComponent] -- ^ multi-line values will be rendered on one line, comma-separated
+  | TopAligned    [StringFormatComponent] -- ^ values will be top-aligned (and bottom-padded to the same height)
   | BottomAligned [StringFormatComponent] -- ^ values will be bottom-aligned (and top-padded)
   deriving (Show, Eq)
 
 data StringFormatComponent =
-    FormatLiteral String        -- ^ Literal text to be rendered as-is
+    FormatLiteral Text          -- ^ Literal text to be rendered as-is
   | FormatField Bool
                 (Maybe Int)
                 (Maybe Int)
@@ -73,6 +77,16 @@ data ReportItemField =
   | FieldNo Int       -- ^ A report item's nth field. May be unimplemented.
     deriving (Show, Eq)
 
+instance Default StringFormat where def = defaultBalanceLineFormat
+
+-- | Default line format for balance report: "%20(total)  %2(depth_spacer)%-(account)"
+defaultBalanceLineFormat :: StringFormat
+defaultBalanceLineFormat = BottomAligned [
+      FormatField False (Just 20) Nothing TotalField
+    , FormatLiteral "  "
+    , FormatField True (Just 2) Nothing DepthSpacerField
+    , FormatField True Nothing Nothing AccountField
+    ]
 ----------------------------------------------------------------------
 
 -- renderStringFormat :: StringFormat -> Map String String -> String
@@ -81,14 +95,14 @@ data ReportItemField =
 ----------------------------------------------------------------------
 
 -- | Parse a string format specification, or return a parse error.
-parseStringFormat :: String -> Either String StringFormat
+parseStringFormat :: Text -> Either String StringFormat
 parseStringFormat input = case (runParser (stringformatp <* eof) "(unknown)") input of
     Left y -> Left $ show y
     Right x -> Right x
 
 defaultStringFormatStyle = BottomAligned
 
-stringformatp :: SimpleStringParser StringFormat
+stringformatp :: SimpleTextParser StringFormat
 stringformatp = do
   alignspec <- optional (try $ char '%' >> oneOf ("^_,"::String))
   let constructor =
@@ -99,19 +113,19 @@ stringformatp = do
           _        -> defaultStringFormatStyle
   constructor <$> many componentp
 
-componentp :: SimpleStringParser StringFormatComponent
+componentp :: SimpleTextParser StringFormatComponent
 componentp = formatliteralp <|> formatfieldp
 
-formatliteralp :: SimpleStringParser StringFormatComponent
+formatliteralp :: SimpleTextParser StringFormatComponent
 formatliteralp = do
-    s <- some c
+    s <- T.pack <$> some c
     return $ FormatLiteral s
     where
       isPrintableButNotPercentage x = isPrint x && x /= '%'
       c =     (satisfy isPrintableButNotPercentage <?> "printable character")
           <|> try (string "%%" >> return '%')
 
-formatfieldp :: SimpleStringParser StringFormatComponent
+formatfieldp :: SimpleTextParser StringFormatComponent
 formatfieldp = do
     char '%'
     leftJustified <- optional (char '-')
@@ -120,13 +134,13 @@ formatfieldp = do
     char '('
     f <- fieldp
     char ')'
-    return $ FormatField (isJust leftJustified) (parseDec minWidth) (parseDec maxWidth) f
+    return $ FormatField (isJust leftJustified) (parseDec minWidth <|> Just 0) (parseDec maxWidth) f
     where
       parseDec s = case s of
         Just text -> Just m where ((m,_):_) = readDec text
         _ -> Nothing
 
-fieldp :: SimpleStringParser ReportItemField
+fieldp :: SimpleTextParser ReportItemField
 fieldp = do
         try (string "account" >> return AccountField)
     <|> try (string "depth_spacer" >> return DepthSpacerField)
@@ -140,12 +154,12 @@ fieldp = do
 formatStringTester fs value expected = actual @?= expected
   where
     actual = case fs of
-      FormatLiteral l                   -> formatString False Nothing Nothing l
-      FormatField leftJustify min max _ -> formatString leftJustify min max value
+      FormatLiteral l                   -> formatText False Nothing Nothing l
+      FormatField leftJustify min max _ -> formatText leftJustify min max value
 
-tests_StringFormat = tests "StringFormat" [
+tests_StringFormat = testGroup "StringFormat" [
 
-   test "formatStringHelper" $ do
+   testCase "formatStringHelper" $ do
       formatStringTester (FormatLiteral " ")                                     ""            " "
       formatStringTester (FormatField False Nothing Nothing DescriptionField)    "description" "description"
       formatStringTester (FormatField False (Just 20) Nothing DescriptionField)  "description" "         description"
@@ -155,25 +169,25 @@ tests_StringFormat = tests "StringFormat" [
       formatStringTester (FormatField True (Just 20) (Just 20) DescriptionField) "description" "description         "
       formatStringTester (FormatField True Nothing (Just 3) DescriptionField)    "description" "des"
 
-  ,let s `gives` expected = test s $ parseStringFormat s @?= Right expected
-   in tests "parseStringFormat" [
+  ,let s `gives` expected = testCase s $ parseStringFormat (T.pack s) @?= Right expected
+   in testGroup "parseStringFormat" [
       ""                           `gives` (defaultStringFormatStyle [])
     , "D"                          `gives` (defaultStringFormatStyle [FormatLiteral "D"])
-    , "%(date)"                    `gives` (defaultStringFormatStyle [FormatField False Nothing Nothing DescriptionField])
-    , "%(total)"                   `gives` (defaultStringFormatStyle [FormatField False Nothing Nothing TotalField])
+    , "%(date)"                    `gives` (defaultStringFormatStyle [FormatField False (Just 0) Nothing DescriptionField])
+    , "%(total)"                   `gives` (defaultStringFormatStyle [FormatField False (Just 0) Nothing TotalField])
     -- TODO
     -- , "^%(total)"                  `gives` (TopAligned [FormatField False Nothing Nothing TotalField])
     -- , "_%(total)"                  `gives` (BottomAligned [FormatField False Nothing Nothing TotalField])
     -- , ",%(total)"                  `gives` (OneLine [FormatField False Nothing Nothing TotalField])
-    , "Hello %(date)!"             `gives` (defaultStringFormatStyle [FormatLiteral "Hello ", FormatField False Nothing Nothing DescriptionField, FormatLiteral "!"])
-    , "%-(date)"                   `gives` (defaultStringFormatStyle [FormatField True Nothing Nothing DescriptionField])
+    , "Hello %(date)!"             `gives` (defaultStringFormatStyle [FormatLiteral "Hello ", FormatField False (Just 0) Nothing DescriptionField, FormatLiteral "!"])
+    , "%-(date)"                   `gives` (defaultStringFormatStyle [FormatField True (Just 0) Nothing DescriptionField])
     , "%20(date)"                  `gives` (defaultStringFormatStyle [FormatField False (Just 20) Nothing DescriptionField])
-    , "%.10(date)"                 `gives` (defaultStringFormatStyle [FormatField False Nothing (Just 10) DescriptionField])
+    , "%.10(date)"                 `gives` (defaultStringFormatStyle [FormatField False (Just 0) (Just 10) DescriptionField])
     , "%20.10(date)"               `gives` (defaultStringFormatStyle [FormatField False (Just 20) (Just 10) DescriptionField])
     , "%20(account) %.10(total)"   `gives` (defaultStringFormatStyle [FormatField False (Just 20) Nothing AccountField
                                                                      ,FormatLiteral " "
-                                                                     ,FormatField False Nothing (Just 10) TotalField
+                                                                     ,FormatField False (Just 0) (Just 10) TotalField
                                                                      ])
-    , test "newline not parsed" $ assertLeft $ parseStringFormat "\n"
+    , testCase "newline not parsed" $ assertLeft $ parseStringFormat "\n"
     ]
  ]
