@@ -14,14 +14,14 @@ module Hledger.Data.Timeclock (
 )
 where
 
-import Data.Maybe
--- import Data.Text (Text)
+import Data.Maybe (fromMaybe)
 import qualified Data.Text as T
-import Data.Time.Calendar
-import Data.Time.Clock
-import Data.Time.Format
-import Data.Time.LocalTime
-import Text.Printf
+import Data.Time.Calendar (addDays)
+import Data.Time.Clock (addUTCTime, getCurrentTime)
+import Data.Time.Format (defaultTimeLocale, formatTime, parseTimeM)
+import Data.Time.LocalTime (LocalTime(..), TimeOfDay(..), getCurrentTimeZone,
+                            localTimeToUTC, midnight, utc, utcToLocalTime)
+import Text.Printf (printf)
 
 import Hledger.Utils
 import Hledger.Data.Types
@@ -77,10 +77,7 @@ timeclockEntriesToTransactions now (i:o:rest)
 {- HLINT ignore timeclockEntriesToTransactions -}
 
 errorExpectedCodeButGot expected actual = errorWithSourceLine line $ "expected timeclock code " ++ (show expected) ++ " but got " ++ show (tlcode actual)
-    where
-        line = case tlsourcepos actual of
-                  GenericSourcePos _ l _ -> l
-                  JournalSourcePos _ (l, _) -> l
+    where line = unPos . sourceLine $ tlsourcepos actual
 
 errorWithSourceLine line msg = error $ "line " ++ show line ++ ": " ++ msg
 
@@ -90,12 +87,12 @@ errorWithSourceLine line msg = error $ "line " ++ show line ++ ": " ++ msg
 entryFromTimeclockInOut :: TimeclockEntry -> TimeclockEntry -> Transaction
 entryFromTimeclockInOut i o
     | otime >= itime = t
-    | otherwise =
-        error' $ "clock-out time less than clock-in time in:\n" ++ showTransaction t  -- PARTIAL:
+    | otherwise = error' . T.unpack $
+        "clock-out time less than clock-in time in:\n" <> showTransaction t  -- PARTIAL:
     where
       t = Transaction {
             tindex       = 0,
-            tsourcepos   = tlsourcepos i,
+            tsourcepos   = (tlsourcepos i, tlsourcepos i),
             tdate        = idate,
             tdate2       = Nothing,
             tstatus      = Cleared,
@@ -116,13 +113,17 @@ entryFromTimeclockInOut i o
       showtime = take 5 . show
       hours    = elapsedSeconds (toutc otime) (toutc itime) / 3600 where toutc = localTimeToUTC utc
       acctname = tlaccount i
-      amount   = Mixed [hrs hours]
+      -- Generate an hours amount. Unusually, we also round the internal Decimal value,
+      -- since otherwise it will often have large recurring decimal parts which (since 1.21)
+      -- print would display all 255 digits of. timeclock amounts have one second resolution,
+      -- so two decimal places is precise enough (#1527).
+      amount   = mixedAmount $ setAmountInternalPrecision 2 $ hrs hours
       ps       = [posting{paccount=acctname, pamount=amount, ptype=VirtualPosting, ptransaction=Just t}]
 
 
 -- tests
 
-tests_Timeclock = tests "Timeclock" [
+tests_Timeclock = testGroup "Timeclock" [
   testCaseSteps "timeclockEntriesToTransactions tests" $ \step -> do
       step "gathering data"
       today <- getCurrentDay
@@ -131,7 +132,7 @@ tests_Timeclock = tests "Timeclock" [
       let now = utcToLocalTime tz now'
           nowstr = showtime now
           yesterday = prevday today
-          clockin = TimeclockEntry nullsourcepos In
+          clockin = TimeclockEntry (initialPos "") In
           mktime d = LocalTime d . fromMaybe midnight .
                      parseTimeM True defaultTimeLocale "%H:%M:%S"
           showtime = formatTime defaultTimeLocale "%H:%M"
