@@ -1,7 +1,6 @@
 {-# LANGUAGE NamedFieldPuns #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE ViewPatterns #-}
-{-# LANGUAGE CPP #-}
 {-|
 
 A 'TransactionModifier' is a rule that modifies certain 'Transaction's,
@@ -13,20 +12,19 @@ module Hledger.Data.TransactionModifier (
 )
 where
 
-import Control.Applicative ((<|>))
-import Data.Maybe
-#if !(MIN_VERSION_base(4,11,0))
-import Data.Monoid ((<>))
-#endif
+import Control.Applicative ((<|>), liftA2)
+import qualified Data.Map as M
+import Data.Maybe (catMaybes)
 import qualified Data.Text as T
-import Data.Time.Calendar
+import Data.Time.Calendar (Day)
 import Hledger.Data.Types
-import Hledger.Data.Dates
 import Hledger.Data.Amount
-import Hledger.Data.Transaction
-import Hledger.Query
-import Hledger.Data.Posting (commentJoin, commentAddTag)
-import Hledger.Utils.Debug
+import Hledger.Data.Dates
+import Hledger.Data.Transaction (txnTieKnot)
+import Hledger.Query (Query, filterQuery, matchesAmount, matchesPostingExtra,
+                      parseQuery, queryIsAmt, queryIsSym, simplifyQuery)
+import Hledger.Data.Posting (commentJoin, commentAddTag, postingAddTags, postingApplyCommodityStyles)
+import Hledger.Utils (dbg6, wrap)
 
 -- $setup
 -- >>> :set -XOverloadedStrings
@@ -38,9 +36,13 @@ import Hledger.Utils.Debug
 -- Or if any of them fails to be parsed, return the first error. A reference
 -- date is provided to help interpret relative dates in transaction modifier
 -- queries.
-modifyTransactions :: Day -> [TransactionModifier] -> [Transaction] -> Either String [Transaction]
-modifyTransactions d tmods ts = do
-  fs <- mapM (transactionModifierToFunction d) tmods  -- convert modifiers to functions, or return a parse error
+modifyTransactions :: (AccountName -> Maybe AccountType)
+                   -> (AccountName -> [Tag])
+                   -> M.Map CommoditySymbol AmountStyle
+                   -> Day -> [TransactionModifier] -> [Transaction]
+                   -> Either String [Transaction]
+modifyTransactions atypes atags styles d tmods ts = do
+  fs <- mapM (transactionModifierToFunction atypes atags styles d) tmods  -- convert modifiers to functions, or return a parse error
   let
     modifytxn t = t''
       where
@@ -61,31 +63,37 @@ modifyTransactions d tmods ts = do
 -- Currently the only kind of modification possible is adding automated
 -- postings when certain other postings are present.
 --
+-- >>> import qualified Data.Text.IO as T
 -- >>> t = nulltransaction{tpostings=["ping" `post` usd 1]}
--- >>> test = either putStr (putStr.showTransaction) . fmap ($ t) . transactionModifierToFunction nulldate
--- >>> test $ TransactionModifier "" ["pong" `post` usd 2]
+-- >>> tmpost acc amt = TMPostingRule (acc `post` amt) False
+-- >>> test = either putStr (T.putStr.showTransaction) . fmap ($ t) . transactionModifierToFunction (const Nothing) (const []) mempty nulldate
+-- >>> test $ TransactionModifier "" ["pong" `tmpost` usd 2]
 -- 0000-01-01
 --     ping           $1.00
 --     pong           $2.00  ; generated-posting: =
 -- <BLANKLINE>
--- >>> test $ TransactionModifier "miss" ["pong" `post` usd 2]
+-- >>> test $ TransactionModifier "miss" ["pong" `tmpost` usd 2]
 -- 0000-01-01
 --     ping           $1.00
 -- <BLANKLINE>
--- >>> test $ TransactionModifier "ping" ["pong" `post` amount{aismultiplier=True, aquantity=3}]
+-- >>> test $ TransactionModifier "ping" [("pong" `tmpost` amount{aquantity=3}){tmprIsMultiplier=True}]
 -- 0000-01-01
 --     ping           $1.00
 --     pong           $3.00  ; generated-posting: = ping
 -- <BLANKLINE>
 --
-transactionModifierToFunction :: Day -> TransactionModifier -> Either String (Transaction -> Transaction)
-transactionModifierToFunction refdate TransactionModifier{tmquerytxt, tmpostingrules} = do
+transactionModifierToFunction :: (AccountName -> Maybe AccountType)
+                              -> (AccountName -> [Tag])
+                              -> M.Map CommoditySymbol AmountStyle
+                              -> Day -> TransactionModifier
+                              -> Either String (Transaction -> Transaction)
+transactionModifierToFunction atypes atags styles refdate TransactionModifier{tmquerytxt, tmpostingrules} = do
   q <- simplifyQuery . fst <$> parseQuery refdate tmquerytxt
   let
-    fs = map (tmPostingRuleToFunction tmquerytxt) tmpostingrules
-    generatePostings ps = [p' | p <- ps
-                              , p' <- if q `matchesPosting` p then p:[f p | f <- fs] else [p]]
-  Right $ \t@(tpostings -> ps) -> txnTieKnot t{tpostings=generatePostings ps}
+    fs = map (\tmpr -> addAccountTags . tmPostingRuleToFunction styles q tmquerytxt tmpr) tmpostingrules
+    addAccountTags p = p `postingAddTags` atags (paccount p)
+    generatePostings p = p : map ($ p) (if matchesPostingExtra atypes q p then fs else [])
+  Right $ \t@(tpostings -> ps) -> txnTieKnot t{tpostings=concatMap generatePostings ps}
 
 -- | Converts a 'TransactionModifier''s posting rule to a 'Posting'-generating function,
 -- which will be used to make a new posting based on the old one (an "automated posting").
@@ -95,9 +103,9 @@ transactionModifierToFunction refdate TransactionModifier{tmquerytxt, tmpostingr
 -- and a hidden _generated-posting: tag which does not.
 -- The TransactionModifier's query text is also provided, and saved
 -- as the tags' value.
-tmPostingRuleToFunction :: T.Text -> TMPostingRule -> (Posting -> Posting)
-tmPostingRuleToFunction querytxt pr =
-  \p -> renderPostingCommentDates $ pr
+tmPostingRuleToFunction :: M.Map CommoditySymbol AmountStyle -> Query -> T.Text -> TMPostingRule -> (Posting -> Posting)
+tmPostingRuleToFunction styles query querytxt tmpr =
+  \p -> postingApplyCommodityStyles styles . renderPostingCommentDates $ pr
       { pdate    = pdate  pr <|> pdate  p
       , pdate2   = pdate2 pr <|> pdate2 p
       , pamount  = amount' p
@@ -107,37 +115,38 @@ tmPostingRuleToFunction querytxt pr =
                    ptags pr
       }
   where
+    pr = tmprPosting tmpr
     qry = "= " <> querytxt
-    amount' = case postingRuleMultiplier pr of
+    symq = filterQuery (liftA2 (||) queryIsSym queryIsAmt) query
+    amount' = case postingRuleMultiplier tmpr of
         Nothing -> const $ pamount pr
         Just n  -> \p ->
           -- Multiply the old posting's amount by the posting rule's multiplier.
           let
-            pramount = dbg6 "pramount" $ head $ amounts $ pamount pr
-            matchedamount = dbg6 "matchedamount" $ pamount p
+            pramount = dbg6 "pramount" . head . amountsRaw $ pamount pr
+            matchedamount = dbg6 "matchedamount" . filterMixedAmount (symq `matchesAmount`) $ pamount p
             -- Handle a matched amount with a total price carefully so as to keep the transaction balanced (#928).
             -- Approach 1: convert to a unit price and increase the display precision slightly
             -- Mixed as = dbg6 "multipliedamount" $ n `multiplyMixedAmount` mixedAmountTotalPriceToUnitPrice matchedamount
             -- Approach 2: multiply the total price (keeping it positive) as well as the quantity
-            Mixed as = dbg6 "multipliedamount" $ n `multiplyMixedAmountAndPrice` matchedamount
+            as = dbg6 "multipliedamount" $ multiplyMixedAmount n matchedamount
           in
             case acommodity pramount of
-              "" -> Mixed as
+              "" -> as
               -- TODO multipliers with commodity symbols are not yet a documented feature.
               -- For now: in addition to multiplying the quantity, it also replaces the
               -- matched amount's commodity, display style, and price with those of the posting rule.
-              c  -> Mixed [a{acommodity = c, astyle = astyle pramount, aprice = aprice pramount} | a <- as]
+              c  -> mapMixedAmount (\a -> a{acommodity = c, astyle = astyle pramount, aprice = aprice pramount}) as
 
 postingRuleMultiplier :: TMPostingRule -> Maybe Quantity
-postingRuleMultiplier p =
-    case amounts $ pamount p of
-        [a] | aismultiplier a -> Just $ aquantity a
-        _                   -> Nothing
+postingRuleMultiplier tmpr = case amountsRaw . pamount $ tmprPosting tmpr of
+    [a] | tmprIsMultiplier tmpr -> Just $ aquantity a
+    _                           -> Nothing
 
 renderPostingCommentDates :: Posting -> Posting
 renderPostingCommentDates p = p { pcomment = comment' }
     where
-        dates = T.concat $ catMaybes [T.pack . showDate <$> pdate p, ("=" <>) . T.pack . showDate <$> pdate2 p]
+        dates = T.concat $ catMaybes [showDate <$> pdate p, ("=" <>) . showDate <$> pdate2 p]
         comment'
             | T.null dates = pcomment p
-            | otherwise    = ("[" <> dates <> "]") `commentJoin` pcomment p
+            | otherwise    = (wrap "[" "]" dates) `commentJoin` pcomment p
