@@ -17,37 +17,55 @@ For more detailed documentation on each type, see the corresponding modules.
 -}
 
 -- {-# LANGUAGE DeriveAnyClass #-}  -- https://hackage.haskell.org/package/deepseq-1.4.4.0/docs/Control-DeepSeq.html#v:rnf
-{-# LANGUAGE DeriveGeneric #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE OverloadedStrings #-}
-{-# LANGUAGE RecordWildCards #-}
-{-# LANGUAGE StandaloneDeriving #-}
-{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE CPP        #-}
+{-# LANGUAGE DeriveGeneric        #-}
+{-# LANGUAGE FlexibleInstances    #-}
+{-# LANGUAGE OverloadedStrings    #-}
+{-# LANGUAGE RecordWildCards      #-}
+{-# LANGUAGE StandaloneDeriving   #-}
 
-module Hledger.Data.Types
+module Hledger.Data.Types (
+  module Hledger.Data.Types,
+#if MIN_VERSION_time(1,11,0)
+  Year
+#endif
+)
 where
 
 import GHC.Generics (Generic)
-import Data.Decimal
-import Data.Default
+import Data.Decimal (Decimal, DecimalRaw(..))
+import Data.Default (Default(..))
 import Data.Functor (($>))
 import Data.List (intercalate)
-import Text.Blaze (ToMarkup(..))
 --XXX https://hackage.haskell.org/package/containers/docs/Data-Map.html
 --Note: You should use Data.Map.Strict instead of this module if:
 --You will eventually need all the values stored.
 --The stored values don't represent large virtual data structures to be lazily computed.
 import qualified Data.Map as M
+import Data.Ord (comparing)
 import Data.Text (Text)
--- import qualified Data.Text as T
-import Data.Time.Calendar
-import Data.Time.LocalTime
+import Data.Time.Calendar (Day)
+import Data.Time.Clock.POSIX (POSIXTime)
+import Data.Time.LocalTime (LocalTime)
 import Data.Word (Word8)
-import System.Time (ClockTime(..))
-import Text.Printf
+import Text.Blaze (ToMarkup(..))
+import Text.Megaparsec (SourcePos)
 
 import Hledger.Utils.Regex
 
+-- synonyms for various date-related scalars
+#if MIN_VERSION_time(1,11,0)
+import Data.Time.Calendar (Year)
+#else
+type Year = Integer
+#endif
+type Month = Int     -- 1-12
+type Quarter = Int   -- 1-4
+type YearWeek = Int  -- 1-52
+type MonthWeek = Int -- 1-5
+type YearDay = Int   -- 1-366
+type MonthDay = Int  -- 1-31
+type WeekDay = Int   -- 1-7
 
 -- | A possibly incomplete year-month-day date provided by the user, to be
 -- interpreted as either a date or a date span depending on context. Missing
@@ -66,10 +84,9 @@ data SmartDate
   = SmartAssumeStart Year (Maybe (Month, Maybe MonthDay))
   | SmartFromReference (Maybe Month) MonthDay
   | SmartMonth Month
-  | SmartRelative SmartSequence SmartInterval
+  | SmartRelative Integer SmartInterval
   deriving (Show)
 
-data SmartSequence = Last | This | Next deriving (Show)
 data SmartInterval = Day | Week | Month | Quarter | Year deriving (Show)
 
 data WhichDate = PrimaryDate | SecondaryDate deriving (Eq,Show)
@@ -78,18 +95,8 @@ data DateSpan = DateSpan (Maybe Day) (Maybe Day) deriving (Eq,Ord,Generic)
 
 instance Default DateSpan where def = DateSpan Nothing Nothing
 
--- synonyms for various date-related scalars
-type Year = Integer
-type Month = Int     -- 1-12
-type Quarter = Int   -- 1-4
-type YearWeek = Int  -- 1-52
-type MonthWeek = Int -- 1-5
-type YearDay = Int   -- 1-366
-type MonthDay = Int  -- 1-31
-type WeekDay = Int   -- 1-7
-
 -- Typical report periods (spans of time), both finite and open-ended.
--- A richer abstraction than DateSpan.
+-- A higher-level abstraction than DateSpan.
 data Period =
     DayPeriod Day
   | WeekPeriod Day
@@ -123,7 +130,7 @@ data Interval =
   | Years Int
   | DayOfMonth Int
   | WeekdayOfMonth Int Int
-  | DayOfWeek Int
+  | DaysOfWeek [Int]
   | DayOfYear Int Int -- Month, Day
   -- WeekOfYear Int
   -- MonthOfYear Int
@@ -132,6 +139,8 @@ data Interval =
 
 instance Default Interval where def = NoInterval
 
+type Payee = Text
+
 type AccountName = Text
 
 data AccountType =
@@ -141,7 +150,31 @@ data AccountType =
   | Revenue
   | Expense
   | Cash  -- ^ a subtype of Asset - liquid assets to show in cashflow report
-  deriving (Show,Eq,Ord,Generic)
+  | Conversion -- ^ a subtype of Equity - account in which to generate conversion postings for transaction prices
+  deriving (Eq,Ord,Generic)
+
+instance Show AccountType where
+  show Asset      = "A"
+  show Liability  = "L"
+  show Equity     = "E"
+  show Revenue    = "R"
+  show Expense    = "X"
+  show Cash       = "C"
+  show Conversion = "V"
+
+-- | Check whether the first argument is a subtype of the second: either equal
+-- or one of the defined subtypes.
+isAccountSubtypeOf :: AccountType -> AccountType -> Bool
+isAccountSubtypeOf Asset      Asset      = True
+isAccountSubtypeOf Liability  Liability  = True
+isAccountSubtypeOf Equity     Equity     = True
+isAccountSubtypeOf Revenue    Revenue    = True
+isAccountSubtypeOf Expense    Expense    = True
+isAccountSubtypeOf Cash       Cash       = True
+isAccountSubtypeOf Cash       Asset      = True
+isAccountSubtypeOf Conversion Conversion = True
+isAccountSubtypeOf Conversion Equity     = True
+isAccountSubtypeOf _          _          = False
 
 -- not worth the trouble, letters defined in accountdirectivep for now
 --instance Read AccountType
@@ -159,6 +192,12 @@ data AccountAlias = BasicAlias AccountName AccountName
 
 data Side = L | R deriving (Eq,Show,Read,Ord,Generic)
 
+-- | One of the decimal marks we support: either period or comma.
+type DecimalMark = Char
+
+isDecimalMark :: Char -> Bool
+isDecimalMark c = c == '.' || c == ','
+
 -- | The basic numeric type used in amounts.
 type Quantity = Decimal
 -- The following is for hledger-web, and requires blaze-markup.
@@ -166,31 +205,38 @@ type Quantity = Decimal
 instance ToMarkup Quantity
  where
    toMarkup = toMarkup . show
+deriving instance Generic (DecimalRaw a)
 
 -- | An amount's per-unit or total cost/selling price in another
 -- commodity, as recorded in the journal entry eg with @ or @@.
 -- Docs call this "transaction price". The amount is always positive.
-data AmountPrice = UnitPrice Amount | TotalPrice Amount
+data AmountPrice = UnitPrice !Amount | TotalPrice !Amount
   deriving (Eq,Ord,Generic,Show)
 
 -- | Display style for an amount.
 data AmountStyle = AmountStyle {
-      ascommodityside   :: Side,                 -- ^ does the symbol appear on the left or the right ?
-      ascommodityspaced :: Bool,                 -- ^ space between symbol and quantity ?
-      asprecision       :: !AmountPrecision,     -- ^ number of digits displayed after the decimal point
-      asdecimalpoint    :: Maybe Char,           -- ^ character used as decimal point: period or comma. Nothing means "unspecified, use default"
-      asdigitgroups     :: Maybe DigitGroupStyle -- ^ style for displaying digit groups, if any
+      ascommodityside   :: !Side,                   -- ^ does the symbol appear on the left or the right ?
+      ascommodityspaced :: !Bool,                   -- ^ space between symbol and quantity ?
+      asprecision       :: !AmountPrecision,        -- ^ number of digits displayed after the decimal point
+      asdecimalpoint    :: !(Maybe Char),           -- ^ character used as decimal point: period or comma. Nothing means "unspecified, use default"
+      asdigitgroups     :: !(Maybe DigitGroupStyle) -- ^ style for displaying digit groups, if any
 } deriving (Eq,Ord,Read,Generic)
 
 instance Show AmountStyle where
-  show AmountStyle{..} =
-    printf "AmountStylePP \"%s %s %s %s %s..\""
-    (show ascommodityside)
-    (show ascommodityspaced)
-    (show asprecision)
-    (show asdecimalpoint)
-    (show asdigitgroups)
+  show AmountStyle{..} = concat
+    [ "AmountStylePP \""
+    , show ascommodityside
+    , show ascommodityspaced
+    , show asprecision
+    , show asdecimalpoint
+    , show asdigitgroups
+    , "..\""
+    ]
 
+-- | The "display precision" for a hledger amount, by which we mean
+-- the number of decimal digits to display to the right of the decimal mark.
+-- This can be from 0 to 255 digits (the maximum supported by the Decimal library),
+-- or NaturalPrecision meaning "show all significant decimal digits".
 data AmountPrecision = Precision !Word8 | NaturalPrecision deriving (Eq,Ord,Read,Show,Generic)
 
 -- | A style for displaying digit groups in the integer part of a
@@ -199,7 +245,7 @@ data AmountPrecision = Precision !Word8 | NaturalPrecision deriving (Eq,Ord,Read
 -- point), and the size of each group, starting with the one nearest
 -- the decimal point. The last group size is assumed to repeat. Eg,
 -- comma between thousands is DigitGroups ',' [3].
-data DigitGroupStyle = DigitGroups Char [Word8]
+data DigitGroupStyle = DigitGroups !Char ![Word8]
   deriving (Eq,Ord,Read,Show,Generic)
 
 type CommoditySymbol = Text
@@ -210,15 +256,64 @@ data Commodity = Commodity {
   } deriving (Show,Eq,Generic) --,Ord)
 
 data Amount = Amount {
-      acommodity  :: CommoditySymbol,   -- commodity symbol, or special value "AUTO"
-      aquantity   :: Quantity,          -- numeric quantity, or zero in case of "AUTO"
-      aismultiplier :: Bool,            -- ^ kludge: a flag marking this amount and posting as a multiplier
-                                        --   in a TMPostingRule. In a regular Posting, should always be false.
-      astyle      :: AmountStyle,
-      aprice      :: Maybe AmountPrice  -- ^ the (fixed, transaction-specific) price for this amount, if any
+      acommodity  :: !CommoditySymbol,     -- commodity symbol, or special value "AUTO"
+      aquantity   :: !Quantity,            -- numeric quantity, or zero in case of "AUTO"
+      astyle      :: !AmountStyle,
+      aprice      :: !(Maybe AmountPrice)  -- ^ the (fixed, transaction-specific) price for this amount, if any
     } deriving (Eq,Ord,Generic,Show)
 
-newtype MixedAmount = Mixed [Amount] deriving (Eq,Ord,Generic,Show)
+newtype MixedAmount = Mixed (M.Map MixedAmountKey Amount) deriving (Generic,Show)
+
+instance Eq  MixedAmount where a == b  = maCompare a b == EQ
+instance Ord MixedAmount where compare = maCompare
+
+-- | Compare two MixedAmounts, substituting 0 for the quantity of any missing
+-- commodities in either.
+maCompare :: MixedAmount -> MixedAmount -> Ordering
+maCompare (Mixed a) (Mixed b) = go (M.toList a) (M.toList b)
+  where
+    go xss@((kx,x):xs) yss@((ky,y):ys) = case compare kx ky of
+                 EQ -> compareQuantities (Just x) (Just y) <> go xs ys
+                 LT -> compareQuantities (Just x) Nothing  <> go xs yss
+                 GT -> compareQuantities Nothing  (Just y) <> go xss ys
+    go ((_,x):xs) [] = compareQuantities (Just x) Nothing  <> go xs []
+    go [] ((_,y):ys) = compareQuantities Nothing  (Just y) <> go [] ys
+    go []         [] = EQ
+    compareQuantities = comparing (maybe 0 aquantity) <> comparing (maybe 0 totalprice)
+    totalprice x = case aprice x of
+                        Just (TotalPrice p) -> aquantity p
+                        _                   -> 0
+
+-- | Stores the CommoditySymbol of the Amount, along with the CommoditySymbol of
+-- the price, and its unit price if being used.
+data MixedAmountKey
+  = MixedAmountKeyNoPrice    !CommoditySymbol
+  | MixedAmountKeyTotalPrice !CommoditySymbol !CommoditySymbol
+  | MixedAmountKeyUnitPrice  !CommoditySymbol !CommoditySymbol !Quantity
+  deriving (Eq,Generic,Show)
+
+-- | We don't auto-derive the Ord instance because it would give an undesired ordering.
+-- We want the keys to be sorted lexicographically:
+-- (1) By the primary commodity of the amount.
+-- (2) By the commodity of the price, with no price being first.
+-- (3) By the unit price, from most negative to most positive, with total prices
+-- before unit prices.
+-- For example, we would like the ordering to give
+-- MixedAmountKeyNoPrice "X" < MixedAmountKeyTotalPrice "X" "Z" < MixedAmountKeyNoPrice "Y"
+instance Ord MixedAmountKey where
+  compare = comparing commodity <> comparing pCommodity <> comparing pPrice
+    where
+      commodity (MixedAmountKeyNoPrice    c)     = c
+      commodity (MixedAmountKeyTotalPrice c _)   = c
+      commodity (MixedAmountKeyUnitPrice  c _ _) = c
+
+      pCommodity (MixedAmountKeyNoPrice    _)      = Nothing
+      pCommodity (MixedAmountKeyTotalPrice _ pc)   = Just pc
+      pCommodity (MixedAmountKeyUnitPrice  _ pc _) = Just pc
+
+      pPrice (MixedAmountKeyNoPrice    _)     = Nothing
+      pPrice (MixedAmountKeyTotalPrice _ _)   = Nothing
+      pPrice (MixedAmountKeyUnitPrice  _ _ q) = Just q
 
 data PostingType = RegularPosting | VirtualPosting | BalancedVirtualPosting
                    deriving (Eq,Show,Generic)
@@ -276,10 +371,10 @@ instance Show Status where -- custom show.. bad idea.. don't do it..
 --    at once. Not implemented, requires #934.
 --
 data BalanceAssertion = BalanceAssertion {
-      baamount    :: Amount,             -- ^ the expected balance in a particular commodity
-      batotal     :: Bool,               -- ^ disallow additional non-asserted commodities ?
-      bainclusive :: Bool,               -- ^ include subaccounts when calculating the actual balance ?
-      baposition  :: GenericSourcePos    -- ^ the assertion's file position, for error reporting
+      baamount    :: Amount,    -- ^ the expected balance in a particular commodity
+      batotal     :: Bool,      -- ^ disallow additional non-asserted commodities ?
+      bainclusive :: Bool,      -- ^ include subaccounts when calculating the actual balance ?
+      baposition  :: SourcePos  -- ^ the assertion's file position, for error reporting
     } deriving (Eq,Generic,Show)
 
 data Posting = Posting {
@@ -290,7 +385,8 @@ data Posting = Posting {
       pamount           :: MixedAmount,
       pcomment          :: Text,              -- ^ this posting's comment lines, as a single non-indented multi-line string
       ptype             :: PostingType,
-      ptags             :: [Tag],                   -- ^ tag names and values, extracted from the comment
+      ptags             :: [Tag],                   -- ^ tag names and values, extracted from the posting comment 
+                                                    --   and (after finalisation) the posting account's directive if any
       pbalanceassertion :: Maybe BalanceAssertion,  -- ^ an expected balance in the account after this posting,
                                                     --   in a single commodity, excluding subaccounts.
       ptransaction      :: Maybe Transaction,       -- ^ this posting's parent transaction (co-recursive types).
@@ -323,20 +419,10 @@ instance Show Posting where
     ,"poriginal="         ++ show poriginal
     ] ++ "}"
 
--- TODO: needs renaming, or removal if no longer needed. See also TextPosition in Hledger.UI.Editor
--- | The position of parse errors (eg), like parsec's SourcePos but generic.
-data GenericSourcePos = GenericSourcePos FilePath Int Int    -- ^ file path, 1-based line number and 1-based column number.
-                      | JournalSourcePos FilePath (Int, Int) -- ^ file path, inclusive range of 1-based line numbers (first, last).
-  deriving (Eq, Read, Show, Ord, Generic)
-
---{-# ANN Transaction "HLint: ignore" #-}
---    Ambiguous type variable ‘p0’ arising from an annotation
---    prevents the constraint ‘(Data p0)’ from being solved.
---    Probable fix: use a type annotation to specify what ‘p0’ should be.
 data Transaction = Transaction {
       tindex                   :: Integer,   -- ^ this transaction's 1-based position in the transaction stream, or 0 when not available
       tprecedingcomment        :: Text,      -- ^ any comment lines immediately preceding this transaction
-      tsourcepos               :: GenericSourcePos,  -- ^ the file position where the date starts
+      tsourcepos               :: (SourcePos, SourcePos),  -- ^ the file position where the date starts, and where the last posting ends
       tdate                    :: Day,
       tdate2                   :: Maybe Day,
       tstatus                  :: Status,
@@ -364,9 +450,12 @@ nulltransactionmodifier = TransactionModifier{
 
 -- | A transaction modifier transformation, which adds an extra posting
 -- to the matched posting's transaction.
--- Can be like a regular posting, or the amount can have the aismultiplier flag set,
+-- Can be like a regular posting, or can have the tmprIsMultiplier flag set,
 -- indicating that it's a multiplier for the matched posting's amount.
-type TMPostingRule = Posting
+data TMPostingRule = TMPostingRule
+  { tmprPosting :: Posting
+  , tmprIsMultiplier :: Bool
+  } deriving (Eq,Generic,Show)
 
 -- | A periodic transaction rule, describing a transaction that recurs.
 data PeriodicTransaction = PeriodicTransaction {
@@ -397,7 +486,7 @@ nullperiodictransaction = PeriodicTransaction{
 data TimeclockCode = SetBalance | SetRequiredHours | In | Out | FinalOut deriving (Eq,Ord,Generic)
 
 data TimeclockEntry = TimeclockEntry {
-      tlsourcepos   :: GenericSourcePos,
+      tlsourcepos   :: SourcePos,
       tlcode        :: TimeclockCode,
       tldatetime    :: LocalTime,
       tlaccount     :: AccountName,
@@ -412,7 +501,6 @@ data PriceDirective = PriceDirective {
   ,pdcommodity :: CommoditySymbol
   ,pdamount    :: Amount
   } deriving (Eq,Ord,Generic,Show)
-        -- Show instance derived in Amount.hs (XXX why ?)
 
 -- | A historical market price (exchange rate) from one commodity to another.
 -- A more concise form of a PriceDirective, without the amount display info.
@@ -421,8 +509,7 @@ data MarketPrice = MarketPrice {
   ,mpfrom :: CommoditySymbol    -- ^ The commodity being converted from.
   ,mpto   :: CommoditySymbol    -- ^ The commodity being converted to.
   ,mprate :: Quantity           -- ^ One unit of the "from" commodity is worth this quantity of the "to" commodity.
-  } deriving (Eq,Ord,Generic)
-        -- Show instance derived in Amount.hs (XXX why ?)
+  } deriving (Eq,Ord,Generic, Show)
 
 -- additional valuation-related types in Valuation.hs
 
@@ -440,16 +527,21 @@ data Journal = Journal {
   -- parsing-related data
    jparsedefaultyear      :: Maybe Year                            -- ^ the current default year, specified by the most recent Y directive (or current date)
   ,jparsedefaultcommodity :: Maybe (CommoditySymbol,AmountStyle)   -- ^ the current default commodity and its format, specified by the most recent D directive
+  ,jparsedecimalmark      :: Maybe DecimalMark                     -- ^ the character to always parse as decimal point, if set by CsvReader's decimal-mark (or a future journal directive)
   ,jparseparentaccounts   :: [AccountName]                         -- ^ the current stack of parent account names, specified by apply account directives
   ,jparsealiases          :: [AccountAlias]                        -- ^ the current account name aliases in effect, specified by alias directives (& options ?)
   -- ,jparsetransactioncount :: Integer                               -- ^ the current count of transactions parsed so far (only journal format txns, currently)
   ,jparsetimeclockentries :: [TimeclockEntry]                       -- ^ timeclock sessions which have not been clocked out
   ,jincludefilestack      :: [FilePath]
   -- principal data
+  ,jdeclaredpayees        :: [(Payee,PayeeDeclarationInfo)]         -- ^ Payees declared by payee directives, in parse order (after journal finalisation)
   ,jdeclaredaccounts      :: [(AccountName,AccountDeclarationInfo)] -- ^ Accounts declared by account directives, in parse order (after journal finalisation)
-  ,jdeclaredaccounttypes  :: M.Map AccountType [AccountName]        -- ^ Accounts whose type has been declared in account directives (usually 5 top-level accounts)
+  ,jdeclaredaccounttags   :: M.Map AccountName [Tag]                -- ^ Accounts which have tags declared in their directives, and those tags. (Does not include parents' tags.)
+  ,jdeclaredaccounttypes  :: M.Map AccountType [AccountName]        -- ^ Accounts whose type has been explicitly declared in their account directives, grouped by type.
+  ,jaccounttypes          :: M.Map AccountName AccountType          -- ^ All accounts for which a type has been declared or can be inferred from its parent or its name.
+  ,jglobalcommoditystyles :: M.Map CommoditySymbol AmountStyle      -- ^ per-commodity display styles declared globally, eg by command line option or import command
   ,jcommodities           :: M.Map CommoditySymbol Commodity        -- ^ commodities and formats declared by commodity directives
-  ,jinferredcommodities   :: M.Map CommoditySymbol AmountStyle      -- ^ commodities and formats inferred from journal amounts  TODO misnamed, should be eg jusedstyles
+  ,jinferredcommodities   :: M.Map CommoditySymbol AmountStyle      -- ^ commodities and formats inferred from journal amounts
   ,jpricedirectives       :: [PriceDirective]                       -- ^ Declarations of market prices by P directives, in parse order (after journal finalisation)
   ,jinferredmarketprices  :: [MarketPrice]                          -- ^ Market prices implied by transactions, in parse order (after journal finalisation)
   ,jtxnmodifiers          :: [TransactionModifier]
@@ -459,11 +551,13 @@ data Journal = Journal {
   ,jfiles                 :: [(FilePath, Text)]                     -- ^ the file path and raw text of the main and
                                                                     --   any included journal files. The main file is first,
                                                                     --   followed by any included files in the order encountered.
-  ,jlastreadtime          :: ClockTime                              -- ^ when this journal was last read from its file(s)
+                                                                    --   TODO: FilePath is a sloppy type here, don't assume it's a
+                                                                    --   real file; values like "", "-", "(string)" can be seen
+  ,jlastreadtime          :: POSIXTime                              -- ^ when this journal was last read from its file(s)
+  -- NOTE: after adding new fields, eg involving account names, consider updating
+  -- the Anon instance in Hleger.Cli.Anon
   } deriving (Eq, Generic)
 
-deriving instance Generic ClockTime
-
 -- | A journal in the process of being parsed, not yet finalised.
 -- The data is partial, and list fields are in reverse order.
 type ParsedJournal = Journal
@@ -472,6 +566,17 @@ type ParsedJournal = Journal
 -- The --output-format option selects one of these for output.
 type StorageFormat = String
 
+-- | Extra information found in a payee directive.
+data PayeeDeclarationInfo = PayeeDeclarationInfo {
+   pdicomment :: Text   -- ^ any comment lines following the payee directive
+  ,pditags    :: [Tag]  -- ^ tags extracted from the comment, if any
+} deriving (Eq,Show,Generic)
+
+nullpayeedeclarationinfo = PayeeDeclarationInfo {
+   pdicomment          = ""
+  ,pditags             = []
+}
+
 -- | Extra information about an account that can be derived from
 -- its account directive (and the other account directives).
 data AccountDeclarationInfo = AccountDeclarationInfo {
@@ -514,7 +619,6 @@ data NormalSign = NormallyPositive | NormallyNegative deriving (Show, Eq)
 -- tree-wise, since each one knows its parent and subs; the first
 -- account is the root of the tree and always exists.
 data Ledger = Ledger {
-  ljournal  :: Journal,
-  laccounts :: [Account]
-}
-
+   ljournal  :: Journal
+  ,laccounts :: [Account]
+  } deriving (Generic)
