@@ -26,15 +26,12 @@ Hledger.Read.Common, to avoid import cycles.
 
 --- ** language
 
-{-# LANGUAGE CPP #-}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE NamedFieldPuns #-}
-{-# LANGUAGE NoMonoLocalBinds #-}
-{-# LANGUAGE OverloadedStrings #-}
-{-# LANGUAGE PackageImports #-}
-{-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE FlexibleContexts    #-}
+{-# LANGUAGE NamedFieldPuns      #-}
+{-# LANGUAGE NoMonoLocalBinds    #-}
+{-# LANGUAGE OverloadedStrings   #-}
+{-# LANGUAGE PackageImports      #-}
 {-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE TupleSections #-}
 
 --- ** exports
 module Hledger.Read.JournalReader (
@@ -42,15 +39,16 @@ module Hledger.Read.JournalReader (
   -- * Reader-finding utils
   findReader,
   splitReaderPrefix,
-  
+
   -- * Reader
   reader,
 
   -- * Parsing utils
-  genericSourcePos,
   parseAndFinaliseJournal,
   runJournalParser,
   rjp,
+  runErroringJournalParser,
+  rejp,
 
   -- * Parsers used elsewhere
   getParentAccount,
@@ -61,7 +59,7 @@ module Hledger.Read.JournalReader (
   datetimep,
   datep,
   modifiedaccountnamep,
-  postingp,
+  tmpostingrulep,
   statusp,
   emptyorcommentlinep,
   followingcommentp,
@@ -73,21 +71,16 @@ module Hledger.Read.JournalReader (
 where
 
 --- ** imports
--- import qualified Prelude (fail)
--- import "base-compat-batteries" Prelude.Compat hiding (fail, readFile)
-import qualified "base-compat-batteries" Control.Monad.Fail.Compat as Fail (fail)
+import qualified Control.Monad.Fail as Fail (fail)
 import qualified Control.Exception as C
-import Control.Monad (forM_, when, void)
+import Control.Monad (forM_, when, void, unless)
 import Control.Monad.IO.Class (MonadIO, liftIO)
 import Control.Monad.Except (ExceptT(..), runExceptT)
-import Control.Monad.State.Strict (get,modify',put)
+import Control.Monad.State.Strict (evalStateT,get,modify',put)
 import Control.Monad.Trans.Class (lift)
 import Data.Char (toLower)
 import Data.Either (isRight)
 import qualified Data.Map.Strict as M
-#if !(MIN_VERSION_base(4,11,0))
-import Data.Semigroup ((<>))
-#endif
 import Data.Text (Text)
 import Data.String
 import Data.List
@@ -114,6 +107,26 @@ import qualified Hledger.Read.CsvReader as CsvReader (reader)
 --- ** doctest setup
 -- $setup
 -- >>> :set -XOverloadedStrings
+--
+--- ** parsing utilities
+
+-- | Run a journal parser in some monad. See also: parseWithState.
+runJournalParser, rjp
+  :: Monad m
+  => JournalParser m a -> Text -> m (Either (ParseErrorBundle Text CustomErr) a)
+runJournalParser p = runParserT (evalStateT p nulljournal) ""
+rjp = runJournalParser
+
+-- | Run an erroring journal parser in some monad. See also: parseWithState.
+runErroringJournalParser, rejp
+  :: Monad m
+  => ErroringJournalParser m a
+  -> Text
+  -> m (Either FinalParseError (Either (ParseErrorBundle Text CustomErr) a))
+runErroringJournalParser p t =
+  runExceptT $ runParserT (evalStateT p nulljournal) "" t
+rejp = runErroringJournalParser
+
 
 --- ** reader finding utilities
 -- Defined here rather than Hledger.Read so that we can use them in includedirectivep below.
@@ -179,11 +192,6 @@ parse iopts = parseAndFinaliseJournal journalp' iopts
       mapM_ addAccountAlias (reverse $ aliasesFromOpts iopts)
       journalp
 
--- | Get the account name aliases from options, if any.
-aliasesFromOpts :: InputOpts -> [AccountAlias]
-aliasesFromOpts = map (\a -> fromparse $ runParser accountaliasp ("--alias "++quoteIfNeeded a) $ T.pack a)
-                  . aliases_
-
 --- ** parsers
 --- *** journal
 
@@ -231,12 +239,14 @@ directivep = (do
    ,applyaccountdirectivep
    ,commoditydirectivep
    ,endapplyaccountdirectivep
+   ,payeedirectivep
    ,tagdirectivep
    ,endtagdirectivep
    ,defaultyeardirectivep
    ,defaultcommoditydirectivep
    ,commodityconversiondirectivep
    ,ignoredpricecommoditydirectivep
+   ,decimalmarkdirectivep
    ]
   ) <?> "directive"
 
@@ -308,6 +318,7 @@ includedirectivep = do
       jparsedefaultyear      = jparsedefaultyear j
       ,jparsedefaultcommodity = jparsedefaultcommodity j
       ,jparseparentaccounts   = jparseparentaccounts j
+      ,jparsedecimalmark      = jparsedecimalmark j
       ,jparsealiases          = jparsealiases j
       ,jcommodities           = jcommodities j
       -- ,jparsetransactioncount = jparsetransactioncount j
@@ -336,12 +347,6 @@ accountdirectivep = do
   -- the account name, possibly modified by preceding alias or apply account directives
   acct <- modifiedaccountnamep
 
-  -- maybe an account type code (ALERX) after two or more spaces
-  -- XXX added in 1.11, deprecated in 1.13, remove in 1.14
-  mtypecode :: Maybe Char <- lift $ optional $ try $ do
-    skipNonNewlineSpaces1 -- at least one more space in addition to the one consumed by modifiedaccountp
-    choice $ map char "ALERX"
-
   -- maybe a comment, on this and/or following lines
   (cmt, tags) <- lift transactioncommentp
 
@@ -351,14 +356,11 @@ accountdirectivep = do
   -- an account type may have been set by account type code or a tag;
   -- the latter takes precedence
   let
-    mtypecode' :: Maybe Text = maybe
-      (T.singleton <$> mtypecode)
-      Just
-      $ lookup accountTypeTagName tags
-    metype = parseAccountTypeCode <$> mtypecode'
+    metype = parseAccountTypeCode <$> lookup accountTypeTagName tags
 
   -- update the journal
   addAccountDeclaration (acct, cmt, tags)
+  unless (null tags) $ addDeclaredAccountTags acct tags
   case metype of
     Nothing         -> return ()
     Just (Right t)  -> addDeclaredAccountType acct t
@@ -370,22 +372,24 @@ accountTypeTagName = "type"
 parseAccountTypeCode :: Text -> Either String AccountType
 parseAccountTypeCode s =
   case T.toLower s of
-    "asset"     -> Right Asset
-    "a"         -> Right Asset
-    "liability" -> Right Liability
-    "l"         -> Right Liability
-    "equity"    -> Right Equity
-    "e"         -> Right Equity
-    "revenue"   -> Right Revenue
-    "r"         -> Right Revenue
-    "expense"   -> Right Expense
-    "x"         -> Right Expense
-    "cash"      -> Right Cash
-    "c"         -> Right Cash
-    _           -> Left err
+    "asset"      -> Right Asset
+    "a"          -> Right Asset
+    "liability"  -> Right Liability
+    "l"          -> Right Liability
+    "equity"     -> Right Equity
+    "e"          -> Right Equity
+    "revenue"    -> Right Revenue
+    "r"          -> Right Revenue
+    "expense"    -> Right Expense
+    "x"          -> Right Expense
+    "cash"       -> Right Cash
+    "c"          -> Right Cash
+    "conversion" -> Right Conversion
+    "v"          -> Right Conversion
+    _            -> Left err
   where
-    err = "invalid account type code "++T.unpack s++", should be one of " ++
-          (intercalate ", " $ ["A","L","E","R","X","C","Asset","Liability","Equity","Revenue","Expense","Cash"])
+    err = T.unpack $ "invalid account type code "<>s<>", should be one of " <>
+            T.intercalate ", " ["A","L","E","R","X","C","V","Asset","Liability","Equity","Revenue","Expense","Cash","Conversion"]
 
 -- Add an account declaration to the journal, auto-numbering it.
 addAccountDeclaration :: (AccountName,Text,[Tag]) -> JournalParser m ()
@@ -401,6 +405,17 @@ addAccountDeclaration (a,cmt,tags) =
              in
                j{jdeclaredaccounts = d:decls})
 
+-- Add a payee declaration to the journal.
+addPayeeDeclaration :: (Payee,Text,[Tag]) -> JournalParser m ()
+addPayeeDeclaration (p, cmt, tags) =
+  modify' (\j@Journal{jdeclaredpayees} -> j{jdeclaredpayees=d:jdeclaredpayees})
+             where
+               d = (p
+                   ,nullpayeedeclarationinfo{
+                     pdicomment = cmt
+                    ,pditags    = tags
+                    })
+
 indentedlinep :: JournalParser m String
 indentedlinep = lift skipNonNewlineSpaces1 >> (rstrip <$> lift restofline)
 
@@ -428,7 +443,7 @@ commoditydirectiveonelinep = do
   lift skipNonNewlineSpaces
   _ <- lift followingcommentp
   let comm = Commodity{csymbol=acommodity, cformat=Just $ dbg6 "style from commodity directive" astyle}
-  if asdecimalpoint astyle == Nothing
+  if isNothing $ asdecimalpoint astyle
   then customFailure $ parseErrorAt off pleaseincludedecimalpoint
   else modify' (\j -> j{jcommodities=M.insert acommodity comm $ jcommodities j})
 
@@ -469,7 +484,7 @@ formatdirectivep expectedsym = do
   _ <- lift followingcommentp
   if acommodity==expectedsym
     then
-      if asdecimalpoint astyle == Nothing
+      if isNothing $ asdecimalpoint astyle
       then customFailure $ parseErrorAt off pleaseincludedecimalpoint
       else return $ dbg6 "style from format subdirective" astyle
     else customFailure $ parseErrorAt off $
@@ -505,34 +520,6 @@ aliasdirectivep = do
   alias <- lift accountaliasp
   addAccountAlias alias
 
-accountaliasp :: TextParser m AccountAlias
-accountaliasp = regexaliasp <|> basicaliasp
-
-basicaliasp :: TextParser m AccountAlias
-basicaliasp = do
-  -- dbgparse 0 "basicaliasp"
-  old <- rstrip <$> (some $ noneOf ("=" :: [Char]))
-  char '='
-  skipNonNewlineSpaces
-  new <- rstrip <$> anySingle `manyTill` eolof  -- eol in journal, eof in command lines, normally
-  return $ BasicAlias (T.pack old) (T.pack new)
-
-regexaliasp :: TextParser m AccountAlias
-regexaliasp = do
-  -- dbgparse 0 "regexaliasp"
-  char '/'
-  off1 <- getOffset
-  re <- some $ noneOf ("/\n\r" :: [Char]) -- paranoid: don't try to read past line end
-  off2 <- getOffset
-  char '/'
-  skipNonNewlineSpaces
-  char '='
-  skipNonNewlineSpaces
-  repl <- anySingle `manyTill` eolof
-  case toRegexCI re of
-    Right r -> return $! RegexAlias r repl
-    Left e  -> customFailure $! parseErrorAtRegion off1 off2 e
-
 endaliasesdirectivep :: JournalParser m ()
 endaliasesdirectivep = do
   keywordsp "end aliases" <?> "end aliases directive"
@@ -552,6 +539,15 @@ endtagdirectivep = do
   lift restofline
   return ()
 
+payeedirectivep :: JournalParser m ()
+payeedirectivep = do
+  string "payee" <?> "payee directive"
+  lift skipNonNewlineSpaces1
+  payee <- lift $ T.strip <$> noncommenttext1p
+  (comment, tags) <- lift transactioncommentp
+  addPayeeDeclaration (payee, comment, tags)
+  return ()
+
 defaultyeardirectivep :: JournalParser m ()
 defaultyeardirectivep = do
   char 'Y' <?> "default year"
@@ -565,7 +561,7 @@ defaultcommoditydirectivep = do
   off <- getOffset
   Amount{acommodity,astyle} <- amountp
   lift restofline
-  if asdecimalpoint astyle == Nothing
+  if isNothing $ asdecimalpoint astyle
   then customFailure $ parseErrorAt off pleaseincludedecimalpoint
   else setDefaultCommodityAndStyle (acommodity, astyle)
 
@@ -601,6 +597,18 @@ commodityconversiondirectivep = do
   lift restofline
   return ()
 
+-- | Read a valid decimal mark from the decimal-mark directive e.g
+--
+-- decimal-mark ,
+decimalmarkdirectivep :: JournalParser m ()
+decimalmarkdirectivep = do
+  string "decimal-mark" <?> "decimal mark"
+  lift skipNonNewlineSpaces1
+  mark <- satisfy isDecimalMark
+  modify' $ \j -> j{jparsedecimalmark=Just mark}
+  lift restofline
+  return ()
+
 --- *** transactions
 
 -- | Parse a transaction modifier (auto postings) rule.
@@ -610,8 +618,8 @@ transactionmodifierp = do
   lift skipNonNewlineSpaces
   querytxt <- lift $ T.strip <$> descriptionp
   (_comment, _tags) <- lift transactioncommentp   -- TODO apply these to modified txns ?
-  postings <- postingsp Nothing
-  return $ TransactionModifier querytxt postings
+  postingrules <- tmpostingrulesp Nothing
+  return $ TransactionModifier querytxt postingrules
 
 -- | Parse a periodic transaction rule.
 --
@@ -638,7 +646,7 @@ periodictransactionp = do
                   Nothing -> today
                   Just y  -> fromGregorian y 1 1
   periodExcerpt <- lift $ excerpt_ $
-                    singlespacedtextsatisfyingp (\c -> c /= ';' && c /= '\n')
+                    singlespacedtextsatisfying1p (\c -> c /= ';' && c /= '\n')
   let periodtxt = T.strip $ getExcerptText periodExcerpt
 
   -- first parsing with 'singlespacedtextp', then "re-parsing" with
@@ -695,7 +703,7 @@ transactionp = do
   let year = first3 $ toGregorian date
   postings <- postingsp (Just year)
   endpos <- getSourcePos
-  let sourcepos = journalSourcePos startpos endpos
+  let sourcepos = (startpos, endpos)
   return $ txnTieKnot $ Transaction 0 "" sourcepos date edate status code description comment tags postings
 
 --- *** postings
@@ -713,61 +721,78 @@ postingsp mTransactionYear = many (postingp mTransactionYear) <?> "postings"
 --   return $ sp ++ (c:cs) ++ "\n"
 
 postingp :: Maybe Year -> JournalParser m Posting
-postingp mTransactionYear = do
-  -- lift $ dbgparse 0 "postingp"
-  (status, account) <- try $ do
-    lift skipNonNewlineSpaces1
-    status <- lift statusp
+postingp = fmap fst . postingphelper False
+
+-- Parse the following whitespace-beginning lines as transaction posting rules, posting
+-- tags, and/or comments (inferring year, if needed, from the given date).
+tmpostingrulesp :: Maybe Year -> JournalParser m [TMPostingRule]
+tmpostingrulesp mTransactionYear = many (tmpostingrulep mTransactionYear) <?> "posting rules"
+
+tmpostingrulep :: Maybe Year -> JournalParser m TMPostingRule
+tmpostingrulep = fmap (uncurry TMPostingRule) . postingphelper True
+
+-- Parse a Posting, and return a flag with whether a multiplier has been detected.
+-- The multiplier is used in TMPostingRules.
+postingphelper :: Bool -> Maybe Year -> JournalParser m (Posting, Bool)
+postingphelper isPostingRule mTransactionYear = do
+    -- lift $ dbgparse 0 "postingp"
+    (status, account) <- try $ do
+      lift skipNonNewlineSpaces1
+      status <- lift statusp
+      lift skipNonNewlineSpaces
+      account <- modifiedaccountnamep
+      return (status, account)
+    let (ptype, account') = (accountNamePostingType account, textUnbracket account)
     lift skipNonNewlineSpaces
-    account <- modifiedaccountnamep
-    return (status, account)
-  let (ptype, account') = (accountNamePostingType account, textUnbracket account)
-  lift skipNonNewlineSpaces
-  amount <- option missingmixedamt $ Mixed . (:[]) <$> amountp
-  lift skipNonNewlineSpaces
-  massertion <- optional balanceassertionp
-  lift skipNonNewlineSpaces
-  (comment,tags,mdate,mdate2) <- lift $ postingcommentp mTransactionYear
-  return posting
-   { pdate=mdate
-   , pdate2=mdate2
-   , pstatus=status
-   , paccount=account'
-   , pamount=amount
-   , pcomment=comment
-   , ptype=ptype
-   , ptags=tags
-   , pbalanceassertion=massertion
-   }
+    mult <- if isPostingRule then multiplierp else pure False
+    amount <- optional $ amountpwithmultiplier mult
+    lift skipNonNewlineSpaces
+    massertion <- optional balanceassertionp
+    lift skipNonNewlineSpaces
+    (comment,tags,mdate,mdate2) <- lift $ postingcommentp mTransactionYear
+    let p = posting
+            { pdate=mdate
+            , pdate2=mdate2
+            , pstatus=status
+            , paccount=account'
+            , pamount=maybe missingmixedamt mixedAmount amount
+            , pcomment=comment
+            , ptype=ptype
+            , ptags=tags
+            , pbalanceassertion=massertion
+            }
+    return (p, mult)
+  where
+    multiplierp = option False $ True <$ char '*'
 
 --- ** tests
 
-tests_JournalReader = tests "JournalReader" [
+tests_JournalReader = testGroup "JournalReader" [
 
    let p = lift accountnamep :: JournalParser IO AccountName in
-   tests "accountnamep" [
-     test "basic" $ assertParse p "a:b:c"
-    -- ,test "empty inner component" $ assertParseError p "a::c" ""  -- TODO
-    -- ,test "empty leading component" $ assertParseError p ":b:c" "x"
-    -- ,test "empty trailing component" $ assertParseError p "a:b:" "x"
+   testGroup "accountnamep" [
+     testCase "basic" $ assertParse p "a:b:c"
+    -- ,testCase "empty inner component" $ assertParseError p "a::c" ""  -- TODO
+    -- ,testCase "empty leading component" $ assertParseError p ":b:c" "x"
+    -- ,testCase "empty trailing component" $ assertParseError p "a:b:" "x"
     ]
 
   -- "Parse a date in YYYY/MM/DD format.
   -- Hyphen (-) and period (.) are also allowed as separators.
   -- The year may be omitted if a default year has been set.
   -- Leading zeroes may be omitted."
-  ,tests "datep" [
-     test "YYYY/MM/DD" $ assertParseEq datep "2018/01/01" (fromGregorian 2018 1 1)
-    ,test "YYYY-MM-DD" $ assertParse datep "2018-01-01"
-    ,test "YYYY.MM.DD" $ assertParse datep "2018.01.01"
-    ,test "yearless date with no default year" $ assertParseError datep "1/1" "current year is unknown"
-    ,test "yearless date with default year" $ do
+  ,testGroup "datep" [
+     testCase "YYYY/MM/DD" $ assertParseEq datep "2018/01/01" (fromGregorian 2018 1 1)
+    ,testCase "YYYY-MM-DD" $ assertParse datep "2018-01-01"
+    ,testCase "YYYY.MM.DD" $ assertParse datep "2018.01.01"
+    ,testCase "yearless date with no default year" $ assertParseError datep "1/1" "current year is unknown"
+    ,testCase "yearless date with default year" $ do
       let s = "1/1"
       ep <- parseWithState nulljournal{jparsedefaultyear=Just 2018} datep s
       either (assertFailure . ("parse error at "++) . customErrorBundlePretty) (const $ return ()) ep
-    ,test "no leading zero" $ assertParse datep "2018/1/1"
+    ,testCase "no leading zero" $ assertParse datep "2018/1/1"
     ]
-  ,test "datetimep" $ do
+  ,testCase "datetimep" $ do
      let
        good = assertParse datetimep
        bad  = (\t -> assertParseError datetimep t "")
@@ -783,9 +808,9 @@ tests_JournalReader = tests "JournalReader" [
      assertParseEq datetimep "2018/1/1 00:00-0800" t
      assertParseEq datetimep "2018/1/1 00:00+1234" t
 
-  ,tests "periodictransactionp" [
+  ,testGroup "periodictransactionp" [
 
-    test "more period text in comment after one space" $ assertParseEq periodictransactionp
+    testCase "more period text in comment after one space" $ assertParseEq periodictransactionp
       "~ monthly from 2018/6 ;In 2019 we will change this\n"
       nullperiodictransaction {
          ptperiodexpr  = "monthly from 2018/6"
@@ -795,7 +820,7 @@ tests_JournalReader = tests "JournalReader" [
         ,ptcomment     = "In 2019 we will change this\n"
         }
 
-    ,test "more period text in description after two spaces" $ assertParseEq periodictransactionp
+    ,testCase "more period text in description after two spaces" $ assertParseEq periodictransactionp
       "~ monthly from 2018/6   In 2019 we will change this\n"
       nullperiodictransaction {
          ptperiodexpr  = "monthly from 2018/6"
@@ -805,7 +830,7 @@ tests_JournalReader = tests "JournalReader" [
         ,ptcomment     = ""
         }
 
-    ,test "Next year in description" $ assertParseEq periodictransactionp
+    ,testCase "Next year in description" $ assertParseEq periodictransactionp
       "~ monthly  Next year blah blah\n"
       nullperiodictransaction {
          ptperiodexpr  = "monthly"
@@ -815,7 +840,7 @@ tests_JournalReader = tests "JournalReader" [
         ,ptcomment     = ""
         }
 
-    ,test "Just date, no description" $ assertParseEq periodictransactionp
+    ,testCase "Just date, no description" $ assertParseEq periodictransactionp
       "~ 2019-01-04\n"
       nullperiodictransaction {
          ptperiodexpr  = "2019-01-04"
@@ -825,74 +850,74 @@ tests_JournalReader = tests "JournalReader" [
         ,ptcomment     = ""
         }
 
-    ,test "Just date, no description + empty transaction comment" $ assertParse periodictransactionp
+    ,testCase "Just date, no description + empty transaction comment" $ assertParse periodictransactionp
       "~ 2019-01-04\n  ;\n  a  1\n  b\n"
 
     ]
 
-  ,tests "postingp" [
-     test "basic" $ assertParseEq (postingp Nothing)
+  ,testGroup "postingp" [
+     testCase "basic" $ assertParseEq (postingp Nothing)
       "  expenses:food:dining  $10.00   ; a: a a \n   ; b: b b \n"
       posting{
         paccount="expenses:food:dining",
-        pamount=Mixed [usd 10],
+        pamount=mixedAmount (usd 10),
         pcomment="a: a a\nb: b b\n",
         ptags=[("a","a a"), ("b","b b")]
         }
 
-    ,test "posting dates" $ assertParseEq (postingp Nothing)
+    ,testCase "posting dates" $ assertParseEq (postingp Nothing)
       " a  1. ; date:2012/11/28, date2=2012/11/29,b:b\n"
       nullposting{
          paccount="a"
-        ,pamount=Mixed [num 1]
+        ,pamount=mixedAmount (num 1)
         ,pcomment="date:2012/11/28, date2=2012/11/29,b:b\n"
         ,ptags=[("date", "2012/11/28"), ("date2=2012/11/29,b", "b")] -- TODO tag name parsed too greedily
         ,pdate=Just $ fromGregorian 2012 11 28
         ,pdate2=Nothing  -- Just $ fromGregorian 2012 11 29
         }
 
-    ,test "posting dates bracket syntax" $ assertParseEq (postingp Nothing)
+    ,testCase "posting dates bracket syntax" $ assertParseEq (postingp Nothing)
       " a  1. ; [2012/11/28=2012/11/29]\n"
       nullposting{
          paccount="a"
-        ,pamount=Mixed [num 1]
+        ,pamount=mixedAmount (num 1)
         ,pcomment="[2012/11/28=2012/11/29]\n"
         ,ptags=[]
         ,pdate= Just $ fromGregorian 2012 11 28
         ,pdate2=Just $ fromGregorian 2012 11 29
         }
 
-    ,test "quoted commodity symbol with digits" $ assertParse (postingp Nothing) "  a  1 \"DE123\"\n"
+    ,testCase "quoted commodity symbol with digits" $ assertParse (postingp Nothing) "  a  1 \"DE123\"\n"
 
-    ,test "only lot price" $ assertParse (postingp Nothing) "  a  1A {1B}\n"
-    ,test "fixed lot price" $ assertParse (postingp Nothing) "  a  1A {=1B}\n"
-    ,test "total lot price" $ assertParse (postingp Nothing) "  a  1A {{1B}}\n"
-    ,test "fixed total lot price, and spaces" $ assertParse (postingp Nothing) "  a  1A {{  =  1B }}\n"
-    ,test "lot price before transaction price" $ assertParse (postingp Nothing) "  a  1A {1B} @ 1B\n"
-    ,test "lot price after transaction price" $ assertParse (postingp Nothing) "  a  1A @ 1B {1B}\n"
-    ,test "lot price after balance assertion not allowed" $ assertParseError (postingp Nothing) "  a  1A @ 1B = 1A {1B}\n" "unexpected '{'"
-    ,test "only lot date" $ assertParse (postingp Nothing) "  a  1A [2000-01-01]\n"
-    ,test "transaction price, lot price, lot date" $ assertParse (postingp Nothing) "  a  1A @ 1B {1B} [2000-01-01]\n"
-    ,test "lot date, lot price, transaction price" $ assertParse (postingp Nothing) "  a  1A [2000-01-01] {1B} @ 1B\n"
+    ,testCase "only lot price" $ assertParse (postingp Nothing) "  a  1A {1B}\n"
+    ,testCase "fixed lot price" $ assertParse (postingp Nothing) "  a  1A {=1B}\n"
+    ,testCase "total lot price" $ assertParse (postingp Nothing) "  a  1A {{1B}}\n"
+    ,testCase "fixed total lot price, and spaces" $ assertParse (postingp Nothing) "  a  1A {{  =  1B }}\n"
+    ,testCase "lot price before transaction price" $ assertParse (postingp Nothing) "  a  1A {1B} @ 1B\n"
+    ,testCase "lot price after transaction price" $ assertParse (postingp Nothing) "  a  1A @ 1B {1B}\n"
+    ,testCase "lot price after balance assertion not allowed" $ assertParseError (postingp Nothing) "  a  1A @ 1B = 1A {1B}\n" "unexpected '{'"
+    ,testCase "only lot date" $ assertParse (postingp Nothing) "  a  1A [2000-01-01]\n"
+    ,testCase "transaction price, lot price, lot date" $ assertParse (postingp Nothing) "  a  1A @ 1B {1B} [2000-01-01]\n"
+    ,testCase "lot date, lot price, transaction price" $ assertParse (postingp Nothing) "  a  1A [2000-01-01] {1B} @ 1B\n"
 
-    ,test "balance assertion over entire contents of account" $ assertParse (postingp Nothing) "  a  $1 == $1\n"
+    ,testCase "balance assertion over entire contents of account" $ assertParse (postingp Nothing) "  a  $1 == $1\n"
     ]
 
-  ,tests "transactionmodifierp" [
+  ,testGroup "transactionmodifierp" [
 
-    test "basic" $ assertParseEq transactionmodifierp
+    testCase "basic" $ assertParseEq transactionmodifierp
       "= (some value expr)\n some:postings  1.\n"
       nulltransactionmodifier {
         tmquerytxt = "(some value expr)"
-       ,tmpostingrules = [nullposting{paccount="some:postings", pamount=Mixed[num 1]}]
+       ,tmpostingrules = [TMPostingRule nullposting{paccount="some:postings", pamount=mixedAmount (num 1)} False]
       }
     ]
 
-  ,tests "transactionp" [
+  ,testGroup "transactionp" [
 
-     test "just a date" $ assertParseEq transactionp "2015/1/1\n" nulltransaction{tdate=fromGregorian 2015 1 1}
+     testCase "just a date" $ assertParseEq transactionp "2015/1/1\n" nulltransaction{tdate=fromGregorian 2015 1 1}
 
-    ,test "more complex" $ assertParseEq transactionp
+    ,testCase "more complex" $ assertParseEq transactionp
       (T.unlines [
         "2012/05/14=2012/05/15 (code) desc  ; tcomment1",
         "    ; tcomment2",
@@ -903,7 +928,7 @@ tests_JournalReader = tests "JournalReader" [
         "    ; ptag2: val2"
         ])
       nulltransaction{
-        tsourcepos=JournalSourcePos "" (1,7),  -- XXX why 7 here ?
+        tsourcepos=(SourcePos "" (mkPos 1) (mkPos 1), SourcePos "" (mkPos 8) (mkPos 1)),  -- 8 because there are 7 lines
         tprecedingcomment="",
         tdate=fromGregorian 2012 5 14,
         tdate2=Just $ fromGregorian 2012 5 15,
@@ -917,7 +942,7 @@ tests_JournalReader = tests "JournalReader" [
             pdate=Nothing,
             pstatus=Cleared,
             paccount="a",
-            pamount=Mixed [usd 1],
+            pamount=mixedAmount (usd 1),
             pcomment="pcomment1\npcomment2\nptag1: val1\nptag2: val2\n",
             ptype=RegularPosting,
             ptags=[("ptag1","val1"),("ptag2","val2")],
@@ -926,7 +951,7 @@ tests_JournalReader = tests "JournalReader" [
           ]
       }
 
-    ,test "parses a well-formed transaction" $
+    ,testCase "parses a well-formed transaction" $
       assertBool "" $ isRight $ rjp transactionp $ T.unlines
         ["2007/01/28 coopportunity"
         ,"    expenses:food:groceries                   $47.18"
@@ -934,10 +959,10 @@ tests_JournalReader = tests "JournalReader" [
         ,""
         ]
 
-    ,test "does not parse a following comment as part of the description" $
+    ,testCase "does not parse a following comment as part of the description" $
       assertParseEqOn transactionp "2009/1/1 a ;comment\n b 1\n" tdescription "a"
 
-    ,test "parses a following whitespace line" $
+    ,testCase "parses a following whitespace line" $
       assertBool "" $ isRight $ rjp transactionp $ T.unlines
         ["2012/1/1"
         ,"  a  1"
@@ -945,7 +970,7 @@ tests_JournalReader = tests "JournalReader" [
         ," "
         ]
 
-    ,test "parses an empty transaction comment following whitespace line" $
+    ,testCase "parses an empty transaction comment following whitespace line" $
       assertBool "" $ isRight $ rjp transactionp $ T.unlines
         ["2012/1/1"
         ,"  ;"
@@ -954,7 +979,7 @@ tests_JournalReader = tests "JournalReader" [
         ," "
         ]
 
-    ,test "comments everywhere, two postings parsed" $
+    ,testCase "comments everywhere, two postings parsed" $
       assertParseEqOn transactionp
         (T.unlines
           ["2009/1/1 x  ; transaction comment"
@@ -970,17 +995,17 @@ tests_JournalReader = tests "JournalReader" [
 
   -- directives
 
-  ,tests "directivep" [
-    test "supports !" $ do
+  ,testGroup "directivep" [
+    testCase "supports !" $ do
         assertParseE directivep "!account a\n"
         assertParseE directivep "!D 1.0\n"
      ]
 
-  ,tests "accountdirectivep" [
-       test "with-comment"       $ assertParse accountdirectivep "account a:b  ; a comment\n"
-      ,test "does-not-support-!" $ assertParseError accountdirectivep "!account a:b\n" ""
-      ,test "account-type-code"  $ assertParse accountdirectivep "account a:b  A\n"
-      ,test "account-type-tag"   $ assertParseStateOn accountdirectivep "account a:b  ; type:asset\n"
+  ,testGroup "accountdirectivep" [
+       testCase "with-comment"       $ assertParse accountdirectivep "account a:b  ; a comment\n"
+      ,testCase "does-not-support-!" $ assertParseError accountdirectivep "!account a:b\n" ""
+      ,testCase "account-type-code"  $ assertParse accountdirectivep "account a:b  ; type:A\n"
+      ,testCase "account-type-tag"   $ assertParseStateOn accountdirectivep "account a:b  ; type:asset\n"
         jdeclaredaccounts
         [("a:b", AccountDeclarationInfo{adicomment          = "type:asset\n"
                                        ,aditags             = [("type","asset")]
@@ -989,28 +1014,28 @@ tests_JournalReader = tests "JournalReader" [
         ]
       ]
 
-  ,test "commodityconversiondirectivep" $ do
+  ,testCase "commodityconversiondirectivep" $ do
      assertParse commodityconversiondirectivep "C 1h = $50.00\n"
 
-  ,test "defaultcommoditydirectivep" $ do
+  ,testCase "defaultcommoditydirectivep" $ do
       assertParse defaultcommoditydirectivep "D $1,000.0\n"
       assertParseError defaultcommoditydirectivep "D $1000\n" "Please include a decimal point or decimal comma"
 
-  ,tests "defaultyeardirectivep" [
-      test "1000" $ assertParse defaultyeardirectivep "Y 1000" -- XXX no \n like the others
-     -- ,test "999" $ assertParseError defaultyeardirectivep "Y 999" "bad year number"
-     ,test "12345" $ assertParse defaultyeardirectivep "Y 12345"
+  ,testGroup "defaultyeardirectivep" [
+      testCase "1000" $ assertParse defaultyeardirectivep "Y 1000" -- XXX no \n like the others
+     -- ,testCase "999" $ assertParseError defaultyeardirectivep "Y 999" "bad year number"
+     ,testCase "12345" $ assertParse defaultyeardirectivep "Y 12345"
      ]
 
-  ,test "ignoredpricecommoditydirectivep" $ do
+  ,testCase "ignoredpricecommoditydirectivep" $ do
      assertParse ignoredpricecommoditydirectivep "N $\n"
 
-  ,tests "includedirectivep" [
-      test "include" $ assertParseErrorE includedirectivep "include nosuchfile\n" "No existing files match pattern: nosuchfile"
-     ,test "glob" $ assertParseErrorE includedirectivep "include nosuchfile*\n" "No existing files match pattern: nosuchfile*"
+  ,testGroup "includedirectivep" [
+      testCase "include" $ assertParseErrorE includedirectivep "include nosuchfile\n" "No existing files match pattern: nosuchfile"
+     ,testCase "glob" $ assertParseErrorE includedirectivep "include nosuchfile*\n" "No existing files match pattern: nosuchfile*"
      ]
 
-  ,test "marketpricedirectivep" $ assertParseEq marketpricedirectivep
+  ,testCase "marketpricedirectivep" $ assertParseEq marketpricedirectivep
     "P 2017/01/30 BTC $922.83\n"
     PriceDirective{
       pddate      = fromGregorian 2017 1 30,
@@ -1018,19 +1043,24 @@ tests_JournalReader = tests "JournalReader" [
       pdamount    = usd 922.83
       }
 
-  ,test "tagdirectivep" $ do
+  ,testGroup "payeedirectivep" [
+       testCase "simple"             $ assertParse payeedirectivep "payee foo\n"
+       ,testCase "with-comment"       $ assertParse payeedirectivep "payee foo ; comment\n"
+       ]
+
+  ,testCase "tagdirectivep" $ do
      assertParse tagdirectivep "tag foo \n"
 
-  ,test "endtagdirectivep" $ do
+  ,testCase "endtagdirectivep" $ do
       assertParse endtagdirectivep "end tag \n"
       assertParse endtagdirectivep "pop \n"
 
-  ,tests "journalp" [
-    test "empty file" $ assertParseEqE journalp "" nulljournal
+  ,testGroup "journalp" [
+    testCase "empty file" $ assertParseEqE journalp "" nulljournal
     ]
 
    -- these are defined here rather than in Common so they can use journalp
-  ,test "parseAndFinaliseJournal" $ do
+  ,testCase "parseAndFinaliseJournal" $ do
       ej <- runExceptT $ parseAndFinaliseJournal journalp definputopts "" "2019-1-1\n"
       let Right j = ej
       assertEqual "" [""] $ journalFilePaths j
