@@ -1,13 +1,21 @@
 {-# LANGUAGE FlexibleContexts, TypeFamilies #-}
-{- | Debugging helpers.
+{- | 
 
-You can enable increasingly verbose debug output by adding --debug [1-9]
-to a hledger command line. --debug with no argument means --debug 1.
-This is implemented by calling dbgN or similar helpers, defined below.
-These calls can be found throughout hledger code; they have been added
-organically where it seemed likely they would be needed again.
-The choice of debug level has not been very systematic.
-202006 Here's a start at some guidelines, not yet applied project-wide:
+Helpers for debug output and pretty-printing 
+(using pretty-simple, with which there may be some overlap).
+This module also exports Debug.Trace.
+
+@dbg0@-@dbg9@ will pretty-print values to stderr
+if the program was run with a sufficiently high @--debug=N@ argument. 
+(@--debug@ with no argument means @--debug=1@; @dbg0@ always prints).
+
+The @debugLevel@ global is set once at startup using unsafePerformIO. 
+In GHCI, this happens only on the first run of :main, so if you want
+to change the debug level without restarting GHCI,
+save a dummy change in Debug.hs and do a :reload.
+(Sometimes it's more convenient to temporarily add dbg0's and :reload.)
+
+In hledger, debug levels are used as follows:
 
 Debug level:  What to show:
 ------------  ---------------------------------------------------------
@@ -22,11 +30,6 @@ Debug level:  What to show:
 8             command line parsing
 9             any other rarely needed / more in-depth info
 
-Tip: when debugging with GHCI, the first run after loading Debug.hs sets the
-debug level. If you need to change it, you must touch Debug.hs, :reload in GHCI,
-then run the command with a new --debug value. Or, often it's more convenient
-to add a temporary dbg0 and :reload (dbg0 always prints).
-
 -}
 
 -- more:
@@ -36,15 +39,20 @@ to add a temporary dbg0 and :reload (dbg0 always prints).
 -- http://hackage.haskell.org/packages/archive/traced/2009.7.20/doc/html/Debug-Traced.html
 
 module Hledger.Utils.Debug (
+  -- * Pretty printing
    pprint
   ,pshow
-  ,ptrace
+  -- * Tracing
   ,traceWith
+  -- * Pretty tracing
+  ,ptrace
+  -- ** Debug-level-aware tracing
+  ,debugLevel
   ,traceAt
   ,traceAtWith
-  ,debugLevel
   ,ptraceAt
   ,ptraceAtWith
+  -- ** Easiest form (recommended)
   ,dbg0
   ,dbg1
   ,dbg2
@@ -55,6 +63,8 @@ module Hledger.Utils.Debug (
   ,dbg7
   ,dbg8
   ,dbg9
+  ,dbgExit
+  -- ** More control
   ,dbg0With
   ,dbg1With
   ,dbg2With
@@ -65,7 +75,7 @@ module Hledger.Utils.Debug (
   ,dbg7With
   ,dbg8With
   ,dbg9With
-  ,dbgExit
+  -- ** For standalone lines in IO blocks
   ,ptraceAtIO
   ,dbg0IO
   ,dbg1IO
@@ -77,37 +87,54 @@ module Hledger.Utils.Debug (
   ,dbg7IO
   ,dbg8IO
   ,dbg9IO
-  ,plog
-  ,plogAt
+  -- ** Trace the state of hledger parsers
   ,traceParse
   ,dbgparse
   ,module Debug.Trace
-)
+  ,useColorOnStdout
+  ,useColorOnStderr
+  ,dlog)
 where
 
 import           Control.Monad (when)
 import           Control.Monad.IO.Class
 import           Data.List hiding (uncons)
 import qualified Data.Text as T
+import qualified Data.Text.Lazy as TL
 import           Debug.Trace
 import           Hledger.Utils.Parse
 import           Safe (readDef)
-import           System.Environment (getArgs)
+import           System.Environment (getArgs, lookupEnv)
 import           System.Exit
 import           System.IO.Unsafe (unsafePerformIO)
 import           Text.Megaparsec
 import           Text.Printf
-import           Text.Show.Pretty (ppShow, pPrint)
+import           Text.Pretty.Simple  -- (defaultOutputOptionsDarkBg, OutputOptions(..), pShowOpt, pPrintOpt)
+import Data.Maybe (isJust)
+import System.Console.ANSI (hSupportsANSIColor)
+import System.IO (stdout, Handle, stderr)
 
--- | Pretty print. Easier alias for pretty-show's pPrint.
+prettyopts = 
+  baseopts
+    { outputOptionsIndentAmount=2
+    , outputOptionsCompact=True
+    }
+  where
+    baseopts
+      | useColorOnStderr = defaultOutputOptionsDarkBg -- defaultOutputOptionsLightBg
+      | otherwise        = defaultOutputOptionsNoColor
+
+-- | Pretty print. Generic alias for pretty-simple's pPrint.
 pprint :: Show a => a -> IO ()
-pprint = pPrint
+pprint = pPrintOpt CheckColorTty prettyopts
 
--- | Pretty show. Easier alias for pretty-show's ppShow.
+-- | Pretty show. Generic alias for pretty-simple's pShow.
 pshow :: Show a => a -> String
-pshow = ppShow
+pshow = TL.unpack . pShowOpt prettyopts
 
--- | Pretty trace. Easier alias for traceShowId + ppShow.
+-- XXX some of the below can be improved with pretty-simple, https://github.com/cdepillabout/pretty-simple#readme
+
+-- | Pretty trace. Easier alias for traceShowId + pShow.
 ptrace :: Show a => a -> a
 ptrace = traceWith pshow
 
@@ -116,20 +143,19 @@ ptrace = traceWith pshow
 traceWith :: Show a => (a -> String) -> a -> a
 traceWith f a = trace (f a) a
 
--- | Global debug level, which controls the verbosity of debug output
--- on the console. The default is 0 meaning no debug output. The
+-- | Global debug level, which controls the verbosity of debug errput
+-- on the console. The default is 0 meaning no debug errput. The
 -- @--debug@ command line flag sets it to 1, or @--debug=N@ sets it to
 -- a higher value (note: not @--debug N@ for some reason).  This uses
 -- unsafePerformIO and can be accessed from anywhere and before normal
 -- command-line processing. When running with :main in GHCI, you must
 -- touch and reload this module to see the effect of a new --debug option.
--- After command-line processing, it is also available as the @debug_@
--- field of 'Hledger.Cli.CliOptions.CliOpts'.
 -- {-# OPTIONS_GHC -fno-cse #-}
--- {-# NOINLINE debugLevel #-}
+{-# NOINLINE debugLevel #-}
+-- Avoid using dbg* in this function (infinite loop).
 debugLevel :: Int
-debugLevel = case snd $ break (=="--debug") args of
-               "--debug":[]  -> 1
+debugLevel = case dropWhile (/="--debug") args of
+               ["--debug"]   -> 1
                "--debug":n:_ -> readDef 1 n
                _             ->
                  case take 1 $ filter ("--debug" `isPrefixOf`) args of
@@ -139,29 +165,130 @@ debugLevel = case snd $ break (=="--debug") args of
     where
       args = unsafePerformIO getArgs
 
+-- Avoid using dbg*, pshow etc. in this function (infinite loop).
+-- | Check the IO environment to see if ANSI colour codes should be used on stdout.
+-- This is done using unsafePerformIO so it can be used anywhere, eg in
+-- low-level debug utilities, which should be ok since we are just reading.
+-- The logic is: use color if
+-- the program was started with --color=yes|always
+-- or (
+--   the program was not started with --color=no|never
+--   and a NO_COLOR environment variable is not defined
+--   and stdout supports ANSI color and -o/--output-file was not used or is "-"
+-- ).
+-- Caveats:
+-- When running code in GHCI, this module must be reloaded to see a change.
+-- {-# OPTIONS_GHC -fno-cse #-}
+-- {-# NOINLINE useColorOnStdout #-}
+useColorOnStdout :: Bool
+useColorOnStdout = not hasOutputFile && useColorOnHandle stdout
+
+-- Avoid using dbg*, pshow etc. in this function (infinite loop).
+-- | Like useColorOnStdout, but checks for ANSI color support on stderr,
+-- and is not affected by -o/--output-file.
+-- {-# OPTIONS_GHC -fno-cse #-}
+-- {-# NOINLINE useColorOnStdout #-}
+useColorOnStderr :: Bool
+useColorOnStderr = useColorOnHandle stderr
+
+-- Avoid using dbg*, pshow etc. in this function (infinite loop).
+-- XXX sorry, I'm just cargo-culting these pragmas:
+-- {-# OPTIONS_GHC -fno-cse #-}
+-- {-# NOINLINE useColorOnHandle #-}
+useColorOnHandle :: Handle -> Bool
+useColorOnHandle h = unsafePerformIO $ do
+  no_color       <- isJust <$> lookupEnv "NO_COLOR"
+  supports_color <- hSupportsANSIColor h
+  let coloroption = colorOption
+  return $ coloroption `elem` ["always","yes"]
+       || (coloroption `notElem` ["never","no"] && not no_color && supports_color)
+
+-- Keep synced with color/colour flag definition in hledger:CliOptions.
+-- Avoid using dbg*, pshow etc. in this function (infinite loop).
+-- | Read the value of the --color or --colour command line option provided at program startup
+-- using unsafePerformIO. If this option was not provided, returns the empty string.
+-- (When running code in GHCI, this module must be reloaded to see a change.)
+-- {-# OPTIONS_GHC -fno-cse #-}
+-- {-# NOINLINE colorOption #-}
+colorOption :: String
+colorOption = 
+  -- similar to debugLevel
+  let args = unsafePerformIO getArgs in
+  case dropWhile (/="--color") args of
+    -- --color ARG
+    "--color":v:_ -> v
+    _ ->
+      case take 1 $ filter ("--color=" `isPrefixOf`) args of
+        -- --color=ARG
+        ['-':'-':'c':'o':'l':'o':'r':'=':v] -> v
+        _ ->
+          case dropWhile (/="--colour") args of
+            -- --colour ARG
+            "--colour":v:_ -> v
+            _ ->
+              case take 1 $ filter ("--colour=" `isPrefixOf`) args of
+                -- --colour=ARG
+                ['-':'-':'c':'o':'l':'o':'u':'r':'=':v] -> v
+                _ -> ""
+
+-- Avoid using dbg*, pshow etc. in this function (infinite loop).
+-- | Check whether the -o/--output-file option has been used at program startup
+-- with an argument other than "-", using unsafePerformIO.
+-- {-# OPTIONS_GHC -fno-cse #-}
+-- {-# NOINLINE hasOutputFile #-}
+hasOutputFile :: Bool
+hasOutputFile = outputFileOption `notElem` [Nothing, Just "-"]
+
+-- Keep synced with output-file flag definition in hledger:CliOptions.
+-- Avoid using dbg*, pshow etc. in this function (infinite loop).
+-- | Read the value of the -o/--output-file command line option provided at program startup,
+-- if any, using unsafePerformIO.
+-- (When running code in GHCI, this module must be reloaded to see a change.)
+-- {-# OPTIONS_GHC -fno-cse #-}
+-- {-# NOINLINE outputFileOption #-}
+outputFileOption :: Maybe String
+outputFileOption = 
+  let args = unsafePerformIO getArgs in
+  case dropWhile (not . ("-o" `isPrefixOf`)) args of
+    -- -oARG
+    ('-':'o':v@(_:_)):_ -> Just v
+    -- -o ARG
+    "-o":v:_ -> Just v
+    _ ->
+      case dropWhile (/="--output-file") args of
+        -- --output-file ARG
+        "--output-file":v:_ -> Just v
+        _ ->
+          case take 1 $ filter ("--output-file=" `isPrefixOf`) args of
+            -- --output=file=ARG
+            ['-':'-':'o':'u':'t':'p':'u':'t':'-':'f':'i':'l':'e':'=':v] -> Just v
+            _ -> Nothing
+
 -- | Trace (print to stderr) a string if the global debug level is at
 -- or above the specified level. At level 0, always prints. Otherwise,
 -- uses unsafePerformIO.
 traceAt :: Int -> String -> a -> a
 traceAt level
-    | level > 0 && debugLevel < level = flip const
+    | level > 0 && debugLevel < level = const id
     | otherwise = trace
 
--- | Trace (print to stderr) a showable value using a custom show function.
-traceAtWith :: (a -> String) -> a -> a
-traceAtWith f a = trace (f a) a
+-- | Trace (print to stderr) a showable value using a custom show function,
+-- if the global debug level is at or above the specified level.
+-- At level 0, always prints. Otherwise, uses unsafePerformIO.
+traceAtWith :: Int -> (a -> String) -> a -> a
+traceAtWith level f a = traceAt level (f a) a
 
 -- | Pretty-print a label and a showable value to the console
 -- if the global debug level is at or above the specified level.
 -- At level 0, always prints. Otherwise, uses unsafePerformIO.
 ptraceAt :: Show a => Int -> String -> a -> a
 ptraceAt level
-    | level > 0 && debugLevel < level = flip const
-    | otherwise = \s a -> let p = ppShow a
+    | level > 0 && debugLevel < level = const id
+    | otherwise = \s a -> let p = pshow a
                               ls = lines p
                               nlorspace | length ls > 1 = "\n"
-                                        | otherwise     = " " ++ take (10 - length s) (repeat ' ')
-                              ls' | length ls > 1 = map (" "++) ls
+                                        | otherwise     = replicate (11 - length s) ' '
+                              ls' | length ls > 1 = map (' ':) ls
                                   | otherwise     = ls
                           in trace (s++":"++nlorspace++intercalate "\n" ls') a
 
@@ -178,6 +305,10 @@ ptraceAtWith level f
                         -- in trace (s++":"++nlorspace++intercalate "\n" ls') a
                         in trace p a
 
+-- | Log a pretty-printed showable value to "./debug.log". Uses unsafePerformIO.
+dlog :: Show a => a -> a
+dlog x = unsafePerformIO $ appendFile "debug.log" (pshow x ++ "\n") >> return x
+
 -- "dbg" would clash with megaparsec.
 -- | Pretty-print a label and the showable value to the console, then return it.
 dbg0 :: Show a => String -> a -> a
@@ -212,6 +343,10 @@ dbg8 = ptraceAt 8
 dbg9 :: Show a => String -> a -> a
 dbg9 = ptraceAt 9
 
+-- | Like dbg0, but also exit the program. Uses unsafePerformIO.
+dbgExit :: Show a => String -> a -> a
+dbgExit msg = const (unsafePerformIO exitFailure) . dbg0 msg
+
 -- | Like dbg0, but takes a custom show function instead of a label.
 dbg0With :: Show a => (a -> String) -> a -> a
 dbg0With = ptraceAtWith 0
@@ -243,10 +378,6 @@ dbg8With = ptraceAtWith 8
 dbg9With :: Show a => (a -> String) -> a -> a
 dbg9With = ptraceAtWith 9
 
--- | Like dbg0, but also exit the program. Uses unsafePerformIO.
-dbgExit :: Show a => String -> a -> a
-dbgExit msg = const (unsafePerformIO exitFailure) . dbg0 msg
-
 -- | Like ptraceAt, but convenient to insert in an IO monad and
 -- enforces monadic sequencing (plus convenience aliases).
 -- XXX These have a bug; they should use
@@ -289,39 +420,8 @@ dbg8IO = ptraceAtIO 8
 dbg9IO :: (MonadIO m, Show a) => String -> a -> m ()
 dbg9IO = ptraceAtIO 9
 
--- | Log a label and a pretty-printed showable value to ./debug.log, then return it.
--- Can fail, see plogAt.
-plog :: Show a => String -> a -> a
-plog = plogAt 0
-
--- | Log a label and a pretty-printed showable value to ./debug.log,
--- if the global debug level is at or above the specified level.
--- At level 0, always logs. Otherwise, uses unsafePerformIO.
--- Tends to fail if called more than once, at least when built with -threaded
--- (Exception: debug.log: openFile: resource busy (file is locked)).
-plogAt :: Show a => Int -> String -> a -> a
-plogAt lvl
-    | lvl > 0 && debugLevel < lvl = flip const
-    | otherwise = \s a ->
-        let p = ppShow a
-            ls = lines p
-            nlorspace | length ls > 1 = "\n"
-                      | otherwise     = " " ++ take (10 - length s) (repeat ' ')
-            ls' | length ls > 1 = map (" "++) ls
-                | otherwise     = ls
-            output = s++":"++nlorspace++intercalate "\n" ls'++"\n"
-        in unsafePerformIO $ appendFile "debug.log" output >> return a
-
--- XXX redundant ? More/less robust than plogAt ?
--- -- | Like dbg, but writes the output to "debug.log" in the current directory.
--- dbglog :: Show a => String -> a -> a
--- dbglog label a =
---   (unsafePerformIO $
---     appendFile "debug.log" $ label ++ ": " ++ ppShow a ++ "\n")
---   `seq` a
-
 -- | Print the provided label (if non-null) and current parser state
--- (position and next input) to the console. (See also megaparsec's dbg.)
+-- (position and next input) to the console. See also megaparsec's dbg.
 traceParse :: String -> TextParser m ()
 traceParse msg = do
   pos <- getSourcePos
@@ -342,5 +442,5 @@ traceParseAt level msg = when (level <= debugLevel) $ traceParse msg
 
 -- | Convenience alias for traceParseAt
 dbgparse :: Int -> String -> TextParser m ()
-dbgparse level msg = traceParseAt level msg
+dbgparse = traceParseAt
 
