@@ -1,7 +1,7 @@
 {-# LANGUAGE FlexibleInstances     #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE OverloadedStrings     #-}
 {-# LANGUAGE ScopedTypeVariables   #-}
-{-# LANGUAGE TypeSynonymInstances  #-}
 {-|
 
 Easy regular expression helpers, currently based on regex-tdfa. These should:
@@ -54,6 +54,7 @@ module Hledger.Utils.Regex (
   ,RegexError
    -- * total regex operations
   ,regexMatch
+  ,regexMatchText
   ,regexReplace
   ,regexReplaceUnmemo
   ,regexReplaceAllBy
@@ -66,6 +67,7 @@ import Data.Array ((!), elems, indices)
 import Data.Char (isDigit)
 import Data.List (foldl')
 import Data.MemoUgly (memo)
+import Data.Text (Text)
 import qualified Data.Text as T
 import Text.Regex.TDFA (
   Regex, CompOption(..), defaultCompOpt, defaultExecOpt,
@@ -73,13 +75,11 @@ import Text.Regex.TDFA (
   RegexLike(..), RegexMaker(..), RegexOptions(..), RegexContext(..)
   )
 
-import Hledger.Utils.UTF8IOCompat (error')
-
 
 -- | Regular expression. Extended regular expression-ish syntax ? But does not support eg (?i) syntax.
 data Regexp
-  = Regexp   { reString :: String, reCompiled :: Regex }
-  | RegexpCI { reString :: String, reCompiled :: Regex }
+  = Regexp   { reString :: Text, reCompiled :: Regex }
+  | RegexpCI { reString :: Text, reCompiled :: Regex }
 
 instance Eq Regexp where
   Regexp   s1 _ == Regexp   s2 _ = s1 == s2
@@ -93,7 +93,7 @@ instance Ord Regexp where
   RegexpCI _ _ `compare` Regexp _ _ = GT
 
 instance Show Regexp where
-  showsPrec d r = showParen (d > app_prec) $ reCons . showsPrec (app_prec+1) (reString r)
+  showsPrec d r = showParen (d > app_prec) $ reCons . showsPrec (app_prec+1) (T.unpack $ reString r)
     where app_prec = 10
           reCons = case r of Regexp   _ _ -> showString "Regexp "
                              RegexpCI _ _ -> showString "RegexpCI "
@@ -108,8 +108,8 @@ instance Read Regexp where
     where app_prec = 10
 
 instance ToJSON Regexp where
-  toJSON (Regexp   s _) = String . T.pack $ "Regexp "   ++ s
-  toJSON (RegexpCI s _) = String . T.pack $ "RegexpCI " ++ s
+  toJSON (Regexp   s _) = String $ "Regexp "   <> s
+  toJSON (RegexpCI s _) = String $ "RegexpCI " <> s
 
 instance RegexLike Regexp String where
   matchOnce = matchOnce . reCompiled
@@ -124,30 +124,32 @@ instance RegexContext Regexp String String where
   matchM = matchM . reCompiled
 
 -- Convert a Regexp string to a compiled Regex, or return an error message.
-toRegex :: String -> Either RegexError Regexp
-toRegex = memo $ \s -> mkRegexErr s (Regexp s <$> makeRegexM s)
+toRegex :: Text -> Either RegexError Regexp
+toRegex = memo $ \s -> mkRegexErr s (Regexp s <$> makeRegexM (T.unpack s))  -- Have to unpack here because Text instance in regex-tdfa only appears in 1.3.1
 
 -- Like toRegex, but make a case-insensitive Regex.
-toRegexCI :: String -> Either RegexError Regexp
-toRegexCI = memo $ \s -> mkRegexErr s (RegexpCI s <$> makeRegexOptsM defaultCompOpt{caseSensitive=False} defaultExecOpt s)
+toRegexCI :: Text -> Either RegexError Regexp
+toRegexCI = memo $ \s -> mkRegexErr s (RegexpCI s <$> makeRegexOptsM defaultCompOpt{caseSensitive=False} defaultExecOpt (T.unpack s))  -- Have to unpack here because Text instance in regex-tdfa only appears in 1.3.1
 
 -- | Make a nice error message for a regexp error.
-mkRegexErr :: String -> Maybe a -> Either RegexError a
+mkRegexErr :: Text -> Maybe a -> Either RegexError a
 mkRegexErr s = maybe (Left errmsg) Right
-  where errmsg = "this regular expression could not be compiled: " ++ s
+  where errmsg = T.unpack $ "this regular expression could not be compiled: " <> s
 
 -- Convert a Regexp string to a compiled Regex, throw an error
-toRegex' :: String -> Regexp
-toRegex' = either error' id . toRegex
+toRegex' :: Text -> Regexp
+toRegex' = either errorWithoutStackTrace id . toRegex
 
 -- Like toRegex', but make a case-insensitive Regex.
-toRegexCI' :: String -> Regexp
-toRegexCI' = either error' id . toRegexCI
+toRegexCI' :: Text -> Regexp
+toRegexCI' = either errorWithoutStackTrace id . toRegexCI
 
 -- | A replacement pattern. May include numeric backreferences (\N).
 type Replacement = String
 
--- | An regular expression compilation/processing error message.
+-- | An error message arising during a regular expression operation.
+-- Eg: trying to compile a malformed regular expression, or
+-- trying to apply a malformed replacement pattern.
 type RegexError = String
 
 -- helpers
@@ -157,18 +159,32 @@ type RegexError = String
 regexMatch :: Regexp -> String -> Bool
 regexMatch = matchTest
 
+-- | Tests whether a Regexp matches a Text.
+--
+-- This currently unpacks the Text to a String an works on that. This is due to
+-- a performance bug in regex-tdfa (#9), which may or may not be relevant here.
+regexMatchText :: Regexp -> Text -> Bool
+regexMatchText r = matchTest r . T.unpack
+
 --------------------------------------------------------------------------------
 -- new total functions
 
 -- | A memoising version of regexReplace. Caches the result for each
 -- search pattern, replacement pattern, target string tuple.
+-- This won't generate a regular expression parsing error since that
+-- is pre-compiled nowadays, but there can still be a runtime error 
+-- from the replacement pattern, eg with a backreference referring 
+-- to a nonexistent match group.
 regexReplace :: Regexp -> Replacement -> String -> Either RegexError String
 regexReplace re repl = memo $ regexReplaceUnmemo re repl
 
 -- helpers:
 
 -- Replace this regular expression with this replacement pattern in this
--- string, or return an error message.
+-- string, or return an error message. (There should be no regexp
+-- parsing errors these days since Regexp's compiled form is used,
+-- but there can still be a runtime error from the replacement
+-- pattern, eg a backreference referring to a nonexistent match group.)
 regexReplaceUnmemo :: Regexp -> Replacement -> String -> Either RegexError String
 regexReplaceUnmemo re repl s = foldM (replaceMatch repl) s (reverse $ match (reCompiled re) s :: [MatchText String])
   where
@@ -176,23 +192,25 @@ regexReplaceUnmemo re repl s = foldM (replaceMatch repl) s (reverse $ match (reC
     -- appropriate for this match. Or return an error message.
     replaceMatch :: Replacement -> String -> MatchText String -> Either RegexError String
     replaceMatch replpat s matchgroups =
-      erepl >>= \repl -> Right $ pre ++ repl ++ post
-      where
-        ((_,(off,len)):_) = elems matchgroups  -- groups should have 0-based indexes, and there should always be at least one, since this is a match
-        (pre, post') = splitAt off s
-        post = drop len post'
-        -- The replacement text: the replacement pattern with all
-        -- numeric backreferences replaced by the appropriate groups
-        -- from this match. Or an error message.
-        erepl = regexReplaceAllByM backrefRegex (lookupMatchGroup matchgroups) replpat
+      case elems matchgroups of 
+        [] -> Right s
+        ((_,(off,len)):_) ->   -- groups should have 0-based indexes, and there should always be at least one, since this is a match
+          erepl >>= \repl -> Right $ pre ++ repl ++ post
           where
-            -- Given some match groups and a numeric backreference,
-            -- return the referenced group text, or an error message.
-            lookupMatchGroup :: MatchText String -> String -> Either RegexError String
-            lookupMatchGroup grps ('\\':s@(_:_)) | all isDigit s =
-              case read s of n | n `elem` indices grps -> Right $ fst (grps ! n)
-                             _                         -> Left $ "no match group exists for backreference \"\\"++s++"\""
-            lookupMatchGroup _ s = Left $ "lookupMatchGroup called on non-numeric-backreference \""++s++"\", shouldn't happen"
+            (pre, post') = splitAt off s
+            post = drop len post'
+            -- The replacement text: the replacement pattern with all
+            -- numeric backreferences replaced by the appropriate groups
+            -- from this match. Or an error message.
+            erepl = regexReplaceAllByM backrefRegex (lookupMatchGroup matchgroups) replpat
+              where
+                -- Given some match groups and a numeric backreference,
+                -- return the referenced group text, or an error message.
+                lookupMatchGroup :: MatchText String -> String -> Either RegexError String
+                lookupMatchGroup grps ('\\':s@(_:_)) | all isDigit s =
+                  case read s of n | n `elem` indices grps -> Right $ fst (grps ! n)  -- PARTIAL: should not fail, all digits
+                                 _                         -> Left $ "no match group exists for backreference \"\\"++s++"\""
+                lookupMatchGroup _ s = Left $ "lookupMatchGroup called on non-numeric-backreference \""++s++"\", shouldn't happen"
     backrefRegex = toRegex' "\\\\[0-9]+"  -- PARTIAL: should not fail
 
 -- regexReplace' :: Regexp -> Replacement -> String -> String
